from multiprocessing import Process, Event
import sqlite3
from tinydb import TinyDB, Query
from tinydb.storages import JSONStorage
from tinydb.middlewares import CachingMiddleware
import os
import json
import shutil

from .individual import Individual


class Timer(Process):
    def __init__(self, interval, function, args=None, kwargs=None):
        if kwargs is None:
            kwargs = {}
        if args is None:
            args = []
        super(Timer, self).__init__(*args, **kwargs)
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = Event()

        self.daemon = True

    def cancel(self):
        self.finished.set()

    def run(self):
        while not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
            self.finished.wait(self.interval)
        self.finished.set()


class DummyDataStore:
    def __init__(self):
        pass

    def sync_individual(self, individuals):
        pass

    def destroy(self):
        pass


class JsonDataStore(DummyDataStore):
    def __init__(self, problem, database_name, mode="write"):
        self.path = os.path.join(os.getcwd(), database_name)
        self.mode = mode
        self.problem = problem

        if mode == "write":
            if not os.path.exists(self.path):
                pass
            else:
                shutil.rmtree(self.path)
            os.makedirs(self.path)

        file_name = os.path.join(self.path, 'problem.json')
        data = json.dumps(problem.to_dict())
        with open(file_name, 'w') as file:
            file.write(data)

    def read(self):
        for filename in os.listdir(self.path):
            if '.json' in filename:
                with open(os.path.join(self.path, filename), 'r') as file:
                    individual = Individual([])
                    individual.__dict__ = json.load(file)
                    population.individuals.append(individual)

    def insert(self, individual):
        file_name = os.path.join(self.path, str(individual.id) + '.json')
        with open(file_name,  'w') as file:
            individual_dict = individual.to_dict()
            data = json.dumps(individual_dict)
            file.write(data)

    def update(self, individual):
        path = ""
        file_name = os.path.join(self.path, str(individual.id) + '.json')
        if os.path.exists(file_name):
            os.remove(file_name)

        if individual.population_id != -1:
            path = os.path.join(self.path, "population_{}".format(individual.population_id))
            if not os.path.exists(path):
                os.mkdir(path)
        else:
            path = self.path
        file_name = os.path.join(path, str(individual.id) + '.json')
        with open(file_name, 'w') as file:
            individual_dict = individual.to_dict()
            data = json.dumps(individual_dict)
            file.write(data)

    def destroy(self):
        pass


class SqliteDataStore(DummyDataStore):
    sql_main_table = "CREATE TABLE IF NOT EXISTS main (name text NOT NULL, description text NOT NULL);"
    sql_parameters_table = "CREATE TABLE IF NOT EXISTS parameters (name text PRIMARY KEY, parameter json not null);"
    sql_costs_table = "CREATE TABLE IF NOT EXISTS costs (name text PRIMARY KEY, cost json not null);"
    sql_individuals_table = "CREATE TABLE IF NOT EXISTS individuals (id int PRIMARY KEY, individual json not null);"

    sql_main_insert = "INSERT INTO main(name, description) VALUES (?,?)"
    sql_parameters_insert = "INSERT INTO parameters(name, parameter) VALUES (?,?)"
    sql_costs_insert = "INSERT INTO costs(name, cost) VALUES (?,?)"

    sql_individuals_upsert = "INSERT INTO individuals (id, individual) VALUES(?,?) ON CONFLICT(id) DO UPDATE SET individual=excluded.individual;"

    sql_main_select = "SELECT * FROM main;"
    sql_parameters_select = "SELECT * FROM parameters;"
    sql_costs_select = "SELECT * FROM costs;"
    sql_individuals_select = "SELECT * FROM individuals;"
    # SELECT json_extract(individual, '$.costs[0]') as cost FROM individuals where cost>10;

    def __init__(self, problem, database_name, mode="write", thread_safe=False):
        self.problem = problem
        self.database_name = database_name
        self.mode = mode
        self.thread_safe = thread_safe
        # cache
        self._conn = None

        if not self.database_name:
            raise RuntimeError("SqliteDataStoreCacheThread: database name is empty.")

        elif self.mode == "write":
            if os.path.exists(database_name):
                os.remove(database_name)

        if self.mode == "write":
            if os.path.exists(self.database_name):
                statinfo = os.stat(self.database_name)
                if statinfo.st_size == 0:
                    self.read_from_datastore()
                else:
                    self._create_structure()
            else:
                self._create_structure()
        elif self.mode == "rewrite":
            self._create_structure()
        elif self.mode == "read":
            self.read_from_datastore()

    def conn(self):
        if self.thread_safe:
            try:
                if self.mode == "write":
                    conn = sqlite3.connect(self.database_name, isolation_level='Exclusive')
                    c = conn.cursor()
                    c.execute('PRAGMA synchronous = 0')
                    c.execute('PRAGMA journal_mode = ON')
                    conn.commit()
                else:
                    conn = sqlite3.connect(self.database_name)
            except sqlite3.Error as e:
                print(e)

            return conn
        else:
            if self._conn is None:
                try:
                    if self.mode == "write":
                        self._conn = sqlite3.connect(self.database_name, isolation_level='Exclusive')
                        c = self._conn.cursor()
                        c.execute('PRAGMA synchronous = 0')
                        c.execute('PRAGMA journal_mode = OFF')
                        self._conn.commit()
                    else:
                        self._conn = sqlite3.connect(self.database_name)
                except sqlite3.Error as e:
                    print(e)

            return self._conn

    def _create_structure(self):
        conn = self.conn()
        c = conn.cursor()

        # structure
        c.execute(self.sql_main_table)
        c.execute(self.sql_costs_table)
        c.execute(self.sql_parameters_table)
        c.execute(self.sql_individuals_table)
        conn.commit()

        # data
        c.execute(self.sql_main_insert, [self.problem.name, self.problem.description])
        for parameter in self.problem.parameters:
            c.execute(self.sql_parameters_insert, [parameter["name"], json.dumps(parameter)])
        for cost in self.problem.costs:
            c.execute(self.sql_costs_insert, [cost["name"], json.dumps(cost)])
        conn.commit()

    def read_from_datastore(self):
        conn = self.conn()
        c = conn.cursor()

        c.execute(self.sql_main_select)
        rows = c.fetchall()

        # main
        self.problem.name = rows[0][0]
        self.problem.description = rows[0][1]

        # parameters
        self.problem.parameters.clear()
        c.execute(self.sql_parameters_select)
        rows = c.fetchall()
        for row in rows:
            parameter = json.loads(row[1])
            self.problem.parameters[parameter["name"]] = parameter

        # costs
        self.problem.costs.clear()
        c.execute(self.sql_costs_select)
        rows = c.fetchall()
        for row in rows:
            cost = json.loads(row[1])
            self.problem.costs[cost["name"]] = cost

        # individuals
        self.problem.individuals.clear()
        c.execute(self.sql_individuals_select)
        rows = c.fetchall()
        for row in rows:
            individual = Individual.from_dict(json.loads(row[1]))
            self.problem.individuals.append(individual)

    def destroy(self):
        if self._conn is not None:
            self._conn.close()

    def sync_individual(self, individual):
        if self.mode == "write" or self.mode == "rewrite":
            conn = self.conn()
            c = conn.cursor()

            # data
            c.execute(self.sql_individuals_upsert, [individual.id, json.dumps(individual.to_dict())])
            conn.commit()


class TinyDataStore(DummyDataStore):
    def __init__(self, problem, database_name, mode="write"):
        self.problem = problem
        self.database_name = database_name
        self.mode = mode

        if not self.database_name:
            raise RuntimeError("TinyDataStoreCacheThread: database name is empty.")

        elif self.mode == "write":
            if os.path.exists(database_name):
                os.remove(database_name)

        if self.mode == "write":
            # self.db = TinyDB(self.database_name)
            # self.db = TinyDB(storage=MemoryStorage)
            self.db = TinyDB(self.database_name, storage=CachingMiddleware(JSONStorage))
            # sort_keys=True, indent=4

            if os.path.exists(self.database_name):
                statinfo = os.stat(self.database_name)
                if statinfo.st_size == 0:
                    # raise RuntimeError("TinyDataStoreCacheThread: database file already exists.
                    # Mode is WRITE (for automatic rewrite you can use REWRITE mode.")
                    self._create_structure()
                else:
                    self.read_from_datastore()
            else:
                self._create_structure()
        elif self.mode == "rewrite":
            self._create_structure()
        elif self.mode == "read":
            self.db = TinyDB(self.database_name)
            self.read_from_datastore()

    def _create_structure(self):
        main = self.db.table('main')
        main.insert({'name': self.problem.name,
                     'description': self.problem.description})
        main.insert({'parameters': self.problem.parameters})
        main.insert({'costs': self.problem.costs})

        individuals = self.db.table('individuals')

    def read_from_datastore(self):
        main = self.db.table('main')
        main_all = main.all()
        self.problem.name = main_all[0]["name"]
        self.problem.description = main_all[0]["description"]
        self.problem.parameters = main_all[1]["parameters"]
        self.problem.costs = main_all[2]["costs"]

        individuals = self.db.table('individuals')
        individuals_all = individuals.all()
        for data in individuals_all:
            individual = Individual()
            self.problem.individuals.append(Individual.from_dict(data))

    def sync_individual(self, individual):
        individuals = self.db.table('individuals')
        individual_query = Query()
        individuals.upsert(individual.to_dict(), individual_query.id == individual.id)

    def destroy(self):
        self.db.close()



import time
from copy import deepcopy

from .algorithm import Algorithm
from .operators import RandomGenerator, SimulatedBinaryCrossover, \
    PmMutator, TournamentSelector, EpsilonDominance, nondominated_truncate, crowding_distance
from .problem import Problem
from .archive import Archive


class GeneralEvolutionaryAlgorithm(Algorithm):
    """ Basis Class for evolutionary algorithms """

    def __init__(self, problem: Problem, name="General Evolutionary Algorithm", evaluator_type=None):
        super().__init__(problem, name, evaluator_type)

        self.parameters_length = len(self.problem.parameters)

        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum evaluations')
        self.options.declare(name='max_population_number', default=100, lower=1,
                             desc='max_population_number')
        self.options.declare(name='max_population_size', default=100, lower=1,
                             desc='Maximal number of individuals in population')

        self.problem = problem

        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)

    def update_particles(self, individuals, offsprings):
        for individual in individuals:
            individual.features = self.features.copy()

    def run(self):
        pass


class GeneticAlgorithm(GeneralEvolutionaryAlgorithm):
    def __init__(self, problem: Problem, name="General Genetic-based Algorithm", evaluator_type=None):
        super().__init__(problem, name, evaluator_type)

        self.generator = None
        self.selector = None
        self.mutator = None
        self.crossover = None

    def generate(self, parents, archive=None):
        offsprings = []
        # deepcopy of parents
        for offspring in parents:
            offspring_copy = deepcopy(offspring)
            offspring_copy.population_id = -1
            offsprings.append(offspring_copy)

        while len(offsprings) < 2 * self.options['max_population_size']:
            parent1 = self.selector.select(parents)

            repeat = True
            while repeat:
                if archive:
                    if len(archive) <= 1:
                        parent2 = self.selector.select(parents)
                    else:
                        parent2 = archive.rand_choice()
                else:
                    parent2 = self.selector.select(parents)

                if parent1 is not parent2:
                    repeat = False

            # crossover
            child1, child2 = self.crossover.cross(parent1, parent2)

            # mutation
            child1 = self.mutator.mutate(child1)
            child2 = self.mutator.mutate(child2)

            # always create new individual
            if not any(child1 == item for item in offsprings) and len(offsprings) < 2 * self.options['max_population_size']:
                child_copy = deepcopy(child1)
                child_copy.population_id = -1
                offsprings.append(child_copy)
            if not any(child2 == item for item in offsprings) and len(offsprings) < 2 * self.options['max_population_size']:
                child_copy = deepcopy(child2)
                child_copy.population_id = -1
                offsprings.append(child_copy)

        return offsprings

    def run(self):
        pass


class NSGAII(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="NSGA_II Evolutionary Algorithm", evaluator_type=None):
        """
         NSGA-II implementation as described in

         [1] K. Deb, A. Pratap, S. Agarwal and T. Meyarivan, "A fast and elitist
             multiobjective genetic algorithm: NSGA-II," in IEEE Transactions on Evolutionary Computation,
             vol. 6, no. 2, pp. 182-197, Apr 2002. doi: 10.1109/4235.996017
        """
        super().__init__(problem, name, evaluator_type)

        self.options.declare(name='prob_cross', default=1.0, lower=0,
                             desc='prob_cross')

        self.options.declare(name='prob_mutation', default=1.0 / (len(problem.parameters)), lower=0,
                             desc='prob_mutation')

        # set random generator
        self.individual_features['dominate'] = set()
        self.individual_features['crowding_distance'] = 0
        self.individual_features['domination_counter'] = 0
        self.individual_features['front_number'] = 0

    def run(self):
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)
        self.generator.init(self.options['max_population_size'])
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])
        self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        self.selector = TournamentSelector(self.problem.parameters)

        # create initial population
        individuals = self.generator.generate()

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        # evaluate
        self.evaluate(individuals)

        # non-dominated sort of individuals
        self.selector.fast_nondominated_sorting(individuals)

        # sync to datastore
        for individual in individuals:
            self.problem.data_store.sync_individual(individual)

        t_s = time.time()
        self.problem.logger.info(
            "NSGA_II: {}/{}".format(self.options['max_population_number'],
                                    self.options['max_population_number'] * self.options['max_population_size']))

        # optimization
        for it in range(self.options['max_population_number']):
            # generate new offsprings
            offsprings = self.generate(individuals)

            # evaluate the offsprings
            self.evaluate(offsprings)

            # make the pareto dominance calculation and calculating the crowding distance
            self.selector.fast_nondominated_sorting(offsprings)

            # truncate
            individuals = nondominated_truncate(offsprings, self.options['max_population_size'])

            for individual in individuals:
                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

        t = time.time() - t_s
        self.problem.logger.info("NSGA_II: elapsed time: {} s".format(t))


class EpsMOEA(GeneticAlgorithm):
    """
    EpsMOEA is proposed by Deb et al[1] which is a efficient algorithm with less computer time because it does not
    contain a sort process for non-dominate set. The searching space are divided into several hyper-boxes under
    box dominance concept, when two solutions in same box, only one solution can survive in the hyperbox, EpsMOEA choose
    the solution nearer to the hyperbox corner.

    .. Ref::

    [1] K. Deb, M. Mohan, S. Mishra, Evaluating the epsilon-domination based multiobjective evolutionary algorithm for
        a quick computation of Pareto-optimal solutions, Evolutionary Computation 13(4) (2005) 501 - 525.
    [2] Siwei Jiang, Zhihua Cai: Enhance the Convergence and Diversity for eps-MOPSO by Uniform Design and
                                 Minimum Reduce Hypervolume
    """

    def __init__(self, problem: Problem, name="EpsMOEA Algorithm", evaluator_type=None):
        super().__init__(problem, name, evaluator_type)

        self.options.declare(name='prob_cross', default=0.6, lower=0,
                             desc='prob_cross')
        self.options.declare(name='prob_mutation', default=0.2, lower=0,
                             desc='prob_mutation'),
        self.options.declare(name='epsilons', default=0.01, lower=1e-6,
                             desc='prob_epsilons')

        self.archive = None

        # set random generator
        self.individual_features['dominate'] = set()
        self.individual_features['crowding_distance'] = 0
        self.individual_features['domination_counter'] = 0
        self.individual_features['front_number'] = 0

    def run(self):
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)
        self.generator.init(self.options['max_population_size'])
        # set crossover
        self.crossover = SimulatedBinaryCrossover(self.problem.parameters, self.options['prob_cross'])
        self.mutator = PmMutator(self.problem.parameters, self.options['prob_mutation'])
        # one individual is selected from the archive and one from the population to create a new individual
        self.selector = TournamentSelector(self.problem.parameters)

        # create initial population
        individuals = self.generator.generate()

        for individual in individuals:
            # append to problem
            self.problem.individuals.append(individual)
            # add to population
            individual.population_id = 0

        # an archive to collect the eps-dominated solutions
        self.archive = Archive(dominance=EpsilonDominance(epsilons=self.options['epsilons']))

        # evaluate individuals
        self.evaluator.evaluate(individuals)

        # archiving the eps-dominating solutions
        for individual in individuals:
            self.archive.add(individual)

        # sync to datastore
        for individual in individuals:
            self.problem.data_store.sync_individual(individual)

        t_s = time.time()
        self.problem.logger.info(
            "Eps-MOEA: {}/{}".format(self.options['max_population_number'], self.options['max_population_size']))

        for it in range(self.options['max_population_number']):
            # generate and evaluate the next generation
            offsprings = self.generate(individuals, archive=self.archive)

            self.evaluator.evaluate(offsprings)

            # pop-acceptance procedure, the dominating offsprings will be preserved in the population and  in the archive
            for individual in offsprings:
                # pareto dominated solutions
                self.selector.pop_acceptance(individuals, individual)
                # archived solutions
                self.archive.add(individual)

                # add to population
                individual.population_id = it + 1
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            # make a new population from the previous population
            # individuals = offsprings

        t = time.time() - t_s
        self.problem.logger.info("Eps-MOEA: {} s".format(t))



from .utils import flatten

from .surrogate import SurrogateModel

import warnings

from scipy.spatial import distance

from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import SGDRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.ensemble import AdaBoostRegressor, GradientBoostingRegressor, RandomForestRegressor, ExtraTreesRegressor, BaggingRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.neighbors import KNeighborsRegressor, RadiusNeighborsRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, Matern, DotProduct, WhiteKernel, ConstantKernel, RationalQuadratic, ExpSineSquared

from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import RandomizedSearchCV


class SurrogateModelScikit(SurrogateModel):
    def __init__(self, problem):
        super().__init__(problem)

        self.train_step = 10
        self.sigma_threshold = 10
        self.score_threshold = 0.5
        self.distance_metric = 'chebyshev'
        self.distance_threshold = 1.0
        self.has_epsilon = False # some regressors has epsilon (GaussianProcessRegressor, ...)

        # stats
        self.score = None
        self.lml = None
        self.lml_gradient = None

    @staticmethod
    def get_mlp_regressor(search=False, verbose=1):
        # Neural network models
        # Multi-layer Perceptron regressor
        gp = MLPRegressor(hidden_layer_sizes=(6), activation='logistic', solver='lbfgs')

        if search:
            parameters = {
                'solver': ['lbfgs'],
                'max_iter': [1000, 1500, 2000],
                'alpha': [1.e-01, 1.e-05, 1.e-09],
                "hidden_layer_sizes": [(3), (10), (20), (50)],
                'random_state': [0, 2, 4, 5, 8, 9],
                "activation": ['logistic', 'relu']
            }

            gp = GridSearchCV(gp, parameters, n_jobs=-1, verbose=verbose)
            # gp = RandomizedSearchCV(gp, parameters, n_jobs=-1, verbose=verbose)

        return gp

    @staticmethod
    def get_extra_trees_regressor(search=False, verbose=1):
        # Ensemble Methods - decision trees
        gp = ExtraTreesRegressor(n_estimators=10)

        if search:
            parameters = {
                'bootstrap': [True],
                'max_depth': [80, 90, 100, 110],
                'max_features': [2],
                'min_samples_leaf': [3, 4, 5],
                'min_samples_split': [8, 10, 12],
                'n_estimators': [10, 50, 100, 500]
            }

            gp = GridSearchCV(gp, parameters, n_jobs=-1, verbose=verbose)

        return gp

    @staticmethod
    def get_random_forest_regressor(search=False, verbose=1):
        # Ensemble Methods - decision trees
        gp = RandomForestRegressor(n_estimators=10)

        if search:
            parameters = {
                'bootstrap': [True],
                'max_depth': [80, 90, 100, 110],
                'max_features': [2],
                'min_samples_leaf': [3, 4, 5],
                'min_samples_split': [8, 10, 12],
                'n_estimators': [10, 50, 100, 500]
            }
            # parameters = {
            #     'bootstrap': [True],
            #     'max_depth': [80, 90],
            #     'max_features': [2],
            #     'min_samples_leaf': [3],
            #     'min_samples_split': [8],
            #     'n_estimators': [10]
            # }

            gp = GridSearchCV(gp, parameters, n_jobs=-1, verbose=verbose)

        return gp

    @staticmethod
    def get_gaussian_process_regressor(search=False, verbose=1):

        ker_rbf1 = RBF(length_scale=1.0, length_scale_bounds=(1e-1, 10.0))
        ker_rbf2 = ConstantKernel(1.0, constant_value_bounds="fixed") * RBF(1.0, length_scale_bounds=(1e-1, 10.0))
        ker_rq = ConstantKernel(1.0, constant_value_bounds="fixed") * RationalQuadratic(alpha=0.1, length_scale=1)
        ker_expsine = ConstantKernel(1.0, constant_value_bounds="fixed") * ExpSineSquared(1.0, 5.0, periodicity_bounds=(1e-2, 1e1))

        kernel_list = [ker_rbf1, ker_rbf2, ker_rq, ker_expsine]

        gp = GaussianProcessRegressor(kernel=RBF(length_scale=1.0, length_scale_bounds=(1e-1, 10.0)))

        if search:
            parameters = {
                "kernel": kernel_list,
                "alpha": [1e-12, 1e-8, 1e-4, 1e-1, 1e1],
                # "alpha": [1e-10],
                "optimizer": ["fmin_l_bfgs_b"],
                "n_restarts_optimizer": [1, 3, 10, 20],
                # "n_restarts_optimizer": [9],
                "normalize_y": [False],
                "copy_X_train": [True],
                "random_state": [0]
            }

            # grid_search = GridSearchCV(gp, param_grid=param_grid, n_jobs=-1, scoring=None) #
            # grid_search = BayesSearchCV(gp, search_spaces=param_grid)
            # grid_search = RandomizedSearchCV(gp, param_distributions=param_grid, n_jobs=-1, n_iter=1, scoring='neg_mean_squared_error')

            gp = GridSearchCV(gp, parameters, n_jobs=-1, verbose=verbose)

        return gp


    """
    # Ensemble Methods
    # AdaBoost regressor
    regressor_test(AdaBoostRegressor(DecisionTreeRegressor(max_depth=4), n_estimators=300))
    # Bagging regressor
    regressor_test(BaggingRegressor(n_estimators=10))
    # Gradient Boosting for regression
    regressor_test(GradientBoostingRegressor(n_estimators=500, max_depth=4, min_samples_split=2, learning_rate=0.01, loss='ls'))
    # Decision Trees
    # Extremely randomized tree regressor
    regressor_test(ExtraTreeRegressor())
    # Decision tree regressor
    regressor_test(DecisionTreeRegressor())
    # Generalized Linear Models
    # regressor_test(SGDRegressor())
    # Nearest Neighbors
    # Regression based on k-nearest neighbors
    regressor_test(KNeighborsRegressor(n_neighbors=6, weights='distance'))
    # Regression based on neighbors within a fixed radius
    regressor_test(RadiusNeighborsRegressor())
    regressor_test(KernelRidge(alpha=1.0))
    regressor_test(SVR(kernel='rbf', C=1e4, gamma=0.1))
    """

    def init_default_regressor(self):
        # default kernel
        kernel = 1.0 * RationalQuadratic(length_scale=1.0, alpha=0.1)
        # default regressor
        self.regressor = GaussianProcessRegressor(kernel=kernel)
        self.has_epsilon = True

    def predict(self, x, *args):
        if self.trained:
            if self.has_epsilon:
                return self.regressor.predict([x], return_std=True)
            else:
                return self.regressor.predict([x])
        else:
            assert 0

    def train(self):
        # disable sklearn warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # print(len(self.x_data), len(self.y_data))
            # print(self.x_data, self.y_data)
            self.regressor.fit(self.x_data, self.y_data)
            # self.regressor.fit(self.x_data, np.ravel(self.y_data, order='C'))
            # self.trained = True

        if self.eval_stats:
            # score
            self.score = self.regressor.score(self.x_data, self.y_data)
            # print("self.score = {} : {}".format(len(self.x_data), self.score))
            # lml (Gaussian regressor)
            if "log_marginal_likelihood" in dir(self.regressor):
                self.lml, self.lml_gradient = self.regressor.log_marginal_likelihood(self.regressor.kernel_.theta, eval_gradient=True)

        # set trained
        self.trained = self.score >= self.score_threshold
        self.trained = True

    def evaluate(self, individual):
        evaluate = True

        if self.trained:
            # disable sklearn warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                if self.has_epsilon:
                    p, sigma = self.predict(individual.vector)
                else:
                    p = self.predict(individual.vector)
                    sigma = 0

            # compute minimal parameter distance
            if self.distance_metric is not None:
                dist = float("inf")
                for parameter in self.x_data:
                    if self.distance_metric == 'chebyshev':
                        dist = min(dist, distance.chebyshev(individual.vector, parameter))
                    else:
                        assert 0

                # print("dist = {}".format(dist))
                # print("sigma = {}".format(sigma))

            # estimated value
            value = flatten(p[0])

            """
            val = self.problem.evaluate(x)
            # self.problem.logger.info("SurrogateGaussianProcess: predict: x: {}, prediction: {}, sigma: {}".format(x, p[0], sigma))
            self.problem.logger.info("SurrogateGaussianProcess: predict: x: {}, prediction: {}, value: {}".format(x, p[0], val))
            value = val
            """

            # increase counter
            self.predict_counter += 1
            evaluate = False

        if evaluate:
            # evaluate problem
            value = self.problem.evaluate(individual)
            # increase counter
            self.eval_counter += 1
            # add training date to surrogate model
            # self.add_data(np.array(x), np.array(value))
            self.add_data(individual.vector, value)

            if self.eval_counter % self.train_step == 0:
                # init default regressor
                if self.regressor is None:
                    self.init_default_regressor()

                # train model
                self.train()

        """
        if self.trained:
            p, sigma = self.predict([x], return_std=True)
            self.problem.logger.info(
                "SurrogateGaussianProcess: predict: x: {}, prediction: {}, sigma: {}".format(x, p[0], sigma))

            if sigma < self.options['sigma_threshold']:
                # estimated value
                value = p[0].tolist()

                # increase counter
                self.predict_counter += 1
                evaluate = False

        if evaluate:
            # evaluate problem
            value = self.problem.evaluate(x)
            # increase counter
            self.eval_counter += 1
            # add training date to surrogate model
            self.add_data(x, value)

            if self.eval_counter % self.options['train_step'] == 0:
                # surrogate model
                counter_eff = 100.0 * self.predict_counter / (
                            self.eval_counter + self.predict_counter)
                # speed up - increase train step
                if counter_eff > 30:
                    self.options['train_step'] = int(self.options['train_step'] * 1.3)
                self.problem.logger.info("Surrogate: learning: {}, predict eff: {}, train_step: {}"
                                  .format(self.predict_counter, counter_eff, self.options['train_step']))

                # init default regressor
                if self.regressor is not None:
                    self.init_default_regressor()

                # train model
                self.train()
        """
        return value




from .benchmark_functions import BenchmarkFunction
from math import cos, pi, sin, sqrt
import numpy as np


class BiObjectiveTestProblem(BenchmarkFunction):
    """
    The goal of this example to show, how we can use Artap to solve a simple,
    bi-objective optimization problem.

    The problem is defined in the following way [GDE3]:

    Minimize f1 = x1
    Minimize f2 = (1+x2) / x1

    subject to
            x1 e [0.1, 1]
            x2 e [0, 5]

    The Pareto - front of the following problem is known, it is a simple
    hyperbola. This problem is very simple for an Evolutionary algorithm, it finds its solution within 20-30 generations.
    NSGA - II algorithm is used to solve this example.

    References
    ----------

    .: [GDE3] The third Evolution Step of Generalized Differential Evolution
    Saku Kukkonen, Jouni Lampinen

    """

    def set(self):
        self.name = 'Biobjective Test Problem'
        self.parameters = [{'name': 'x_1', 'bounds': [0.1, 1.]},
                           {'name': 'x_2', 'bounds': [0.0, 5.0]}]

        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def pareto_front(self, x):
        """ x -> y """
        x = x.vector
        return 1. / x[0]

    def evaluate(self, individual):
        # The individual.vector function contains the problem parameters in the appropriate (previously defined) order
        f1 = individual.vector[0]
        f2 = (1 + individual.vector[1]) / individual.vector[0]
        return [f1, f2]


class PoloniFunction(BenchmarkFunction):
    """
    Poloni (POL) function is a classic test function for multi-objective optimization methods as its Pareto optimal
    solutions are not continous and non-convex.

    References:
    ---------

    [1] Lei, Gang; Zhu, Jianguo; Guo, Youguang, "Design Optimization Methods for Electrical Machines", 2016

    [2] Lei G, Shao KR, Guo YG, Zhu JG (2012), "Multiobjective sequential optimization method for the design of
        industrial electromagnetic devices. IEEE Trans Magn 48(11):4538–4541"

    section: 4.4.2
    formula: 4.13 and 3.13
    """

    def set(self, **kwargs):
        self.name = 'Booth Function'

        self.set_dimension = 2.0
        self.parameters = [{'name': 'x', 'bounds': [-pi, pi]},
                           {'name': 'y', 'bounds': [-pi, pi]}]

        self.global_optimum = 0.0
        self.global_optimum_coords = [1., 3.]
        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def evaluate(self, x):
        x1 = x.vector[0]
        x2 = x.vector[1]

        A1 = 0.5 * np.sin(1.) - 2. * np.cos(1.) + np.sin(2.) - 1.5 * np.cos(2.)
        A2 = 1.5 * np.sin(1.) - np.cos(1.) + 2. * np.sin(2.) - 0.5 * np.cos(2.)

        B1 = 0.5 * np.sin(x1) - 2. * np.cos(x1) + np.sin(x2) - 1.5 * np.cos(x2)
        B2 = 1.5 * np.sin(x1) - np.cos(x1) + 2. * np.sin(x2) - 0.5 * np.cos(x2)

        f1 = 1. + (A1 - B1) ** 2. + (A2 - B2) ** 2.
        f2 = (x1 + 3.) ** 2. + (x2 + 1.) ** 2.

        return [f1, f2]


class DTLZI(BenchmarkFunction):
    """
    The number of dimensions should be n = m + k -1, where m is the number of the objectives and
    k is an arbitrary number, which should be bigger than 1.

    .. math::

        \left.\matrix{ {\rm Minimize} & f_{1}({\rm x})={1\over 2}x_{1}x_{2}\cdots x_{M-1}(1+g({\rm x}_{M})),
        \hfill\cr {\rm Minimize} & f_{2}({\rm x})={1\over 2}x_{1}x_{2}\cdots (1-x_{M-1})(1+g({\rm x}_{M})),
        \hfill\cr \vdots & \vdots\hfill \cr {\rm Minimize} & f_{M-1}({\rm x})={1\over 2}x_{1}(1-x_{2})(1+g({\rm x}_{M})),
        \hfill\cr {\rm Minimize}& f_{M}({\rm x})={1\over 2}(1-x_{1})(1+g({\rm x}_{M})),
        \hfill\cr {\rm Subject}\,{\rm to}& 0\leq x_{i}, \leq 1, \quad {\rm for}
        i=1,2, \ldots, n.\hfill}\right\}\eqno{\hbox{(7)}}

        References
        ----------

        [1] Deb, K., Thiele, L., Laumanns, M., & Zitzler, E. (2002, May).
            Scalable multi-objective optimization test problems.
            In Proceedings of the 2002 Congress on Evolutionary Computation.
            CEC'02 (Cat. No. 02TH8600) (Vol. 1, pp. 825-830). IEEE.
    """

    def set(self, **kwargs):
        self.name = 'DTLZ I Test Problem'

        self.set_dimension(**kwargs)
        self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)

        self.global_optimum = [0.5 for x in range(self.dimension)]  ###
        self.global_optimum_coords = [0.5 for x in range(self.dimension)]

        # single objective problem
        self.costs = self.generate_objective_functions(**kwargs)

    def pareto_front(self, x):
        """ x -> y """

        p_f = []
        for i in range(0, len(self.costs)):
            p_f.append(0.5 - x)
        return p_f

    def evaluate(self, x):
        x = x.vector
        m = len(self.costs)
        nvar = len(x)
        k = nvar - m + 1  # k >= 1, it can be k = 5 is the offered value,

        g = sum([(y - 0.5) * (y - 0.5) - cos(20.0 * pi * (y - 0.5))
                 for y in x[nvar - k:]])

        # for i in range(len(x) - k, len(x)):
        #     gm += (x[i] - 0.5) ** 2. - cos(20. * pi * (x[i] - 0.5))

        g = 100 * (k + g)

        scores = []
        factor = 0.5 * (1 + g)

        for i in range(0, m):
            fi = factor
            for j in range(0, m - i - 1):
                fi *= x[j]

            if i > 0:
                fi *= (1. - x[m - i - 1])

            scores.append(fi)

        return scores


class DTLZII(BenchmarkFunction):
    """
    The number of dimensions should be n = m + k -1, where m is the number of the objectives and
    k is an arbitrary number, which should be bigger than 1.

    .. math::
        \matrix{ {\rm Minimize} & fi({\rm x})=(1+g({\rm x}_{M}))\cos(x_{1}\pi/2)\cdots\cos(x_{M-1}\pi/2),
        \hfill\cr {\rm Minimize} & f_{2}({\rm x})=(1+g({\rm x}_{M}))\cos(x_{1}\pi/2)\cdots\sin(x_{M-1}\pi/2),
        \hfill\cr \vdots & \vdots \hfill\cr {\rm Minimize} & f_{M}({\rm x})=(1+g({\rm x}_{M}))\sin(x_{1}\pi/2),
        \hfill\cr & 0\leq x_{i}\leq 1,\quad {\rm for} i=1,2, \ldots, n,
        \hfill\cr {\rm where} & g({\rm x}_{M})=\sum_{x_{i}\in {\rm x}_{M}}(x_{i}-0.5)^{2}.\hfill}\eqno{\hbox{(9)}}


        References
        ----------

        [1] Deb, K., Thiele, L., Laumanns, M., & Zitzler, E. (2002, May).
            Scalable multi-objective optimization test problems.
            In Proceedings of the 2002 Congress on Evolutionary Computation.
            CEC'02 (Cat. No. 02TH8600) (Vol. 1, pp. 825-830). IEEE.
    """

    def set(self, **kwargs):
        self.name = 'DTLZ II Test Problem'

        self.set_dimension(**kwargs)
        self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)

        self.global_optimum = [0.5 for x in range(self.dimension)]  ###
        self.global_optimum_coords = [0.5 for x in range(self.dimension)]

        # single objective problem
        self.costs = self.generate_objective_functions(**kwargs)

    def pareto_front(self, x):
        """ x -> y """

        p_f = []
        for i in range(0, len(self.costs)):
            p_f.append(0.5)
        return p_f

    def evaluate(self, x):

        k = 10  # k >= 1, it can be k = 5 is the offered value,

        m = len(self.costs)
        x = x.vector
        scores = []
        for i in range(0, m):
            fi = 1.0
            for j in range(0, m - i - 1):
                fi *= cos(0.5 * x[j] * pi)

            if i > 0:
                fi *= sin(x[m - i] * pi / 2.)
            gm = 0.
            for i in range(0, k):
                gm += (x[len(x) - i - 1] - 0.5) ** 2.
            fi *= (1. + gm)
            scores.append(fi)

        return scores


class DTLZIII(BenchmarkFunction):
    """
    This is the same function like DTLZ II, with the g function of the DTLZ I problem.

    The number of dimensions should be n = m + k -1, where m is the number of the objectives and
    k is an arbitrary number, which should be bigger than 1.


        References
        ----------

        [1] Deb, K., Thiele, L., Laumanns, M., & Zitzler, E. (2002, May).
            Scalable multi-objective optimization test problems.
            In Proceedings of the 2002 Congress on Evolutionary Computation.
            CEC'02 (Cat. No. 02TH8600) (Vol. 1, pp. 825-830). IEEE.
    """

    def set(self, **kwargs):
        self.name = 'DTLZ III Test Problem'

        self.set_dimension(**kwargs)
        self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)

        self.global_optimum = [0.5 for x in range(self.dimension)]  ###
        self.global_optimum_coords = [0.5 for x in range(self.dimension)]

        # single objective problem
        self.costs = self.generate_objective_functions(**kwargs)

    def pareto_front(self, x):
        """ x -> y """

        p_f = []
        for i in range(0, len(self.costs)):
            p_f.append(0.5)
        return p_f

    def evaluate(self, x):

        k = 10  # k >= 1, it can be k = 5 is the offered value,

        m = len(self.costs)
        x = x.vector
        scores = []
        for i in range(0, m):
            fi = 1.0
            for j in range(0, m - i - 1):
                fi *= cos(0.5 * x[j] * pi)

            if i > 0:
                fi *= sin(x[m - i] * pi / 2.)
            # gm = 0.
            # for i in range(0, k):
            #     gm += (x[len(x) - i-1] - 0.5) ** 2.
            # fi *= (1. +  gm)
            # scores.append(fi)
            # g(xm)
            gm = float(k)
            for i in range(0, k):
                gm += (x[len(x) - i - 1] - 0.5) ** 2. - cos(20. * pi * (x[len(x) - i - 1] - 0.5))
            fi = fi * (1 + 100. * gm)
            scores.append(fi)

        return scores


class DTLZIV(BenchmarkFunction):
    """
    xi -> xi**alpha, where alpha = 100

        References
        ----------

        [1] Deb, K., Thiele, L., Laumanns, M., & Zitzler, E. (2002, May).
            Scalable multi-objective optimization test problems.
            In Proceedings of the 2002 Congress on Evolutionary Computation.
            CEC'02 (Cat. No. 02TH8600) (Vol. 1, pp. 825-830). IEEE.
    """

    def set(self, **kwargs):
        self.name = 'DTLZ IV Test Problem'

        self.set_dimension(**kwargs)
        self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)

        self.global_optimum = [0.5 for x in range(self.dimension)]  ###
        self.global_optimum_coords = [0.5 for x in range(self.dimension)]

        # single objective problem
        self.costs = self.generate_objective_functions(**kwargs)

    def pareto_front(self, x):
        """ x -> y """

        p_f = []
        for i in range(0, len(self.costs)):
            p_f.append(0.5)
        return p_f

    def evaluate(self, x):

        k = 10  # k >= 1, it can be k = 5 is the offered value,

        alpha = 100
        m = len(self.costs)
        x = x.vector
        scores = []
        for i in range(0, m):
            fi = 1.0
            for j in range(0, m - i - 1):
                fi *= cos(0.5 * x[j] ** alpha * pi)

            if i > 0:
                fi *= sin(x[m - i] ** alpha * pi / 2.)
            gm = 0.
            for i in range(0, k):
                gm += (x[len(x) - i - 1] - 0.5) ** 2.
            fi *= (1. + gm)
            scores.append(fi)

        return scores


# TODO : finish the dtlz 5-7 problems and the dtlz test problems https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.zdt1


class CEC2020MMF1(BenchmarkFunction):
    """
    Properties
    =================================================================
    Scalable number of variables :        x
    Scalable number of objectives:        x
    Pareto optima known:                  yes
    Pareto front geometry:                convex
    Pareto set geometry:                  nonlinear
    Scalable number of Pareto set:        x
    Nr of global optimums:                2
    NR of local optimums:                 1
    =================================================================

    Solution
    --------
    global pareto set:

    x1 = x1
    x2 = sin(6*pi*abs(x1-2)+pi)

    global pareto front:    f2 =  1 - sqrt(f1) , where 0<=f1<=1

    :param x1: [1,3]
    :param x2: [-1,1]
    :return: [f1, f2]
    """

    def set(self, **kwargs):
        self.name = 'CEC2020 -- MMF1 --'

        self.set_dimension = 2.0
        self.parameters = [{'name': 'x1', 'bounds': [1.0, 3.0]},
                           {'name': 'x2', 'bounds': [-1.0, 1.0]}]

        # self.global_optimum = 0.0
        # self.global_optimum_coords = [1., 3.]
        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def evaluate(self, x):
        x1 = x.vector[0]
        x2 = x.vector[1]

        a = np.where(x1 < 2., 1., np.e)
        f1 = np.abs(x1 - 2)
        f2 = 1 - np.abs(x1 - 2) ** 0.5 + 2. * (x2 - (a ** x1) * np.sin(6. * np.pi * np.abs(x1 - 2.) + np.pi)) ** 2.
        return [f1, f2]


class CEC2020MMF2(BenchmarkFunction):
    """
    Properties
    =================================================================
    Scalable number of variables :        x
    Scalable number of objectives:        x
    Pareto optima known:                  yes
    Pareto front geometry:                convex
    Pareto set geometry:                  nonlinear
    Scalable number of Pareto set:        x
    Nr of global optimums:                2
    NR of local optimums:                 1
    =================================================================

    Solution
    --------
    global pareto set:

    x1 = x2**2  if x1<1
         (x2-1) if 1<x1<2
    x2 = x2

    global pareto front:    f2 =  1 - sqrt(f1) , where 0<=f1<=1

    :param x1: [0,1]
    :param x2: [0,2]
    :return: [f1, f2]
    """

    def set(self, **kwargs):
        self.name = 'CEC2020 -- MMF2 --'

        self.set_dimension = 2.0
        self.parameters = [{'name': 'x1', 'bounds': [0.0, 1.0]},
                           {'name': 'x2', 'bounds': [0.0, 2.0]}]

        # self.global_optimum = 0.0
        # self.global_optimum_coords = [1., 3.]
        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def evaluate(self, x):
        x1 = x.vector[0]
        x2 = x.vector[1]

        f1 = np.abs(x1)

        a = x1 ** 0.5

        f2a = 1 - a + 2. * (4. * (x2 - a)) ** 2. - 2. * np.cos(20.0 * np.pi / (2.0 ** 0.5) * (x2 - a)) + 2.
        f2b = 1 - a + 2. * (4. * (x2 - 1.0 - a)) ** 2. - 2. * np.cos(20.0 * np.pi / (2.0 ** 0.5) * (x2 - 1.0 - a)) + 2.

        f2 = np.where(x2 < 1., f2a, f2b)

        return [f1, f2]


class ZDT1(BenchmarkFunction):
    """
    ZDT1 Test function -- The pareto front should be convex ~ sqrt(x[0])

    Zitzler, Eckart, Kalyanmoy Deb, and Lothar Thiele. “Comparison of multiobjective evolutionary algorithms:
    Empirical results.” Evolutionary computation 8.2 (2000): 173-195. doi: 10.1.1.30.5848
    """

    def set(self, **kwargs):
        self.name = 'ZDT1 I Test Problem'

        self.dimension = 30
        self.parameters = self.generate_paramlist(self.dimension, lb=0.0, ub=1.0)

        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def pareto_front(self, ref_x):
        """
        The exact value should be 1-sqrt(x) on [0,1]
        :param ref_x: list of the solutions (x-coordinates)
        :return: list of [x,y] values
        """

        """ Value of 1-sqrt(x) on the last result of the pareto calculation """
        # the resulting values of the calculateon
        l_sol = []
        for elem in ref_x:
            l_sol.append((elem, 1. - elem ** 0.5))
        return l_sol

    def evaluate(self, x):
        g = self.eval_g(x)
        h = self.eval_h(x.vector[0], g)

        f1 = x.vector[0]
        f2 = h * g

        return [f1, f2]

    def eval_g(self, x):
        g = sum(x.vector) - x.vector[0]
        constant = 9.0 / (len(x.vector) - 1)

        return constant * g + 1.0

    def eval_h(self, f: float, g: float) -> float:
        return 1.0 - sqrt(f / g)


if __name__ == '__main__':
    # visualcheck
    test = PoloniFunction()
    test.plot_2d()

    test = CEC2020MMF1()
    test.plot_2d(f=0)
    test.plot_2d(f=1)

    test = CEC2020MMF2()
    test.plot_2d(f=0)
    test.plot_2d(f=1)



"""
import os
import unittest
import tempfile
from shutil import copyfile, rmtree
from artap.problem import ProblemSqliteDataStore
from artap.datastore import SqliteDataStore
from artap.results import GraphicalResults

from skimage.io import imread
from skimage.measure import compare_mse as mse


class TestDataStore(unittest.TestCase):
    def setUp(self):
        self.working_dir = tempfile.mkdtemp()
        copyfile("." + os.sep + "workspace" + os.sep + "common_data" + os.sep + "data.sqlite",
                 self.working_dir + os.sep + "data.sqlite")
        database_name = self.working_dir + os.sep + "data.sqlite"
        problem = ProblemSqliteDataStore(database_name=database_name)

        self.results = GraphicalResults(problem)

    def tearDown(self):
        rmtree(self.working_dir)
        pass

    def test_scatter(self):
        self.results.plot_scatter('F_1', 'F_2', filename=self.working_dir + os.sep + "scatter.png")

        original = imread("." + os.sep + "workspace" + os.sep + "common_data" + os.sep + "scatter.png")
        picture = imread(self.working_dir + os.sep + "scatter.png")

        mmse = mse(original, picture)
        self.assertLess(mmse, 100)

    def test_individuals(self):
        self.results.plot_individuals('F_1', filename=self.working_dir + os.sep + "individuals.png")

        original = imread("." + os.sep + "workspace" + os.sep + "common_data" + os.sep + "individuals.png")
        picture = imread(self.working_dir + os.sep + "individuals.png")

        mmse = mse(original, picture)
        self.assertLess(mmse, 100)


if __name__ == '__main__':
    unittest.main()
"""


from abc import abstractmethod, ABC
import sys
import random
import math
from math import exp
import numpy as np
import functools
from .individual import Individual
from .utils import VectorAndNumbers
from .doe import build_box_behnken, build_lhs, build_full_fact, build_plackett_burman
from .job import Job
from joblib import Parallel, delayed
from copy import deepcopy

EPSILON = sys.float_info.epsilon

# global constants to p and q dominates
QDOM = 1
PDOM = 2


class Operator(ABC):

    def __init__(self):
        pass

    @staticmethod
    def clip(value, min_value, max_value):
        return max(min_value, min(value, max_value))


# TODO: Change name of classes to Evaluator?
class Evaluator(Operator):

    def __init__(self, algorithm):
        super().__init__()
        self.algorithm = algorithm
        self.individuals = []
        self.job = Job(self.algorithm.problem)

    def add(self, individual):
        self.individuals.append(individual)

    def run(self):
        self.evaluate(self.individuals)

    def evaluate(self, individuals):
        if self.algorithm.options["max_processes"] > 1:
            self.evaluate_parallel(individuals)
        else:
            self.evaluate_serial(individuals)

    def evaluate_serial(self, individuals: list):
        for individual in individuals:
            if individual.state == individual.State.EMPTY:
                individual.costs.append(self.job.evaluate(individual))

    def evaluate_parallel(self, individuals: list):
        # simple parallel loop
        Parallel(n_jobs=self.algorithm.options["max_processes"], verbose=1, require='sharedmem')(
            delayed(self.job.evaluate)(individual)
            for individual in individuals)

    def evaluate_scalar(self, vector):
        individual = Individual(list(vector))

        # add to problem
        self.algorithm.problem.individuals.append(individual)

        self.job.evaluate(individual)
        return individual.costs[0]


class GradientEvaluator(Evaluator):

    def __init__(self, algorithm):
        super().__init__(algorithm)
        self.delta = 1e-4
        self.to_evaluate = []
        self.algorithm.problem.costs.append({'name': 'sensitivity', 'criteria': 'minimize'})
        self.n = len(algorithm.problem.costs)

    def add(self, individual):
        self.individuals.append(individual)
        individual.children = []

        for i in range(len(individual.vector)):
            vector = individual.vector.copy()
            vector[i] += self.delta
            individual.children.append(Individual(vector))
            individual.children[-1].parents.append(individual)

        self.to_evaluate.append(individual)
        self.to_evaluate.extend(individual.children)

    def evaluate(self, individuals):
        for individual in individuals:
            self.add(individual)
        self.run()

    def evaluate_scalar(self, x):
        individual = Individual(x)

        self.add(individual)
        self.run()
        self.algorithm.problem.individuals.append(individual)
        self.to_evaluate = []

        return individual.costs[0]

    def run(self):
        n_params = len(self.individuals[0].vector)
        super().evaluate(self.to_evaluate)
        for individual in self.individuals:
            gradient = np.zeros(n_params)
            i = 0
            for child in individual.children:
                gradient[i] = ((child.costs[0] - individual.costs[0]) / self.delta)
                i += 1
            individual.features['gradient'] = gradient
            if any(gradient < 0):
                sensitivity = 1000
            else:
                sensitivity = sum(abs(gradient))
            if len(individual.costs) > self.n:
                individual.costs[-1] = sensitivity
                individual.costs_signed[-2] = sensitivity
            else:
                individual.costs.append(sensitivity)
                individual.costs_signed.insert(-1, sensitivity)

        self.individuals = []
        self.to_evaluate = []


class WorstCaseEvaluator(Evaluator):

    def __init__(self, algorithm):
        super().__init__(algorithm)
        self.algorithm.problem.costs.append({'name': 'sensitivity', 'criteria': 'minimize'})
        self.n = len(algorithm.problem.costs)
        self.to_evaluate = []

    def add(self, individual):
        parameters = self.algorithm.problem.parameters
        individual.children = []
        self.individuals.append(individual)
        for i in range(len(individual.vector)):
            parameter = parameters[i]
            for sign in [-1, 1]:
                vector = individual.vector.copy()
                vector[i] += sign * parameter['tol']
                individual.children.append(Individual(vector))
                individual.children[-1].parents.append(individual)

        self.to_evaluate.append(individual)
        self.to_evaluate.extend(individual.children)

    def evaluate(self, individuals):
        super().evaluate(individuals)
        for individual in individuals:
            self.add(individual)
        self.run()

    def evaluate_scalar(self, x):
        parent_individual = Individual(x)
        parent_individual.costs.append(self.job.evaluate(parent_individual))
        self.add(parent_individual)
        for individual in self.to_evaluate:
            individual.costs.append(self.job.evaluate(individual))
        self.algorithm.problem.individuals.append(parent_individual)
        self.to_evaluate = []
        return parent_individual.costs[0]

    def run(self):
        super().evaluate(self.to_evaluate)
        for individual in self.individuals:
            sensitivity = []
            for child in individual.children:
                sensitivity.append(abs(individual.costs[0] - child.costs[0]))
            individual.features['sensitivity'] = sum(sensitivity)

            if len(individual.costs) > self.n:
                individual.costs[-1] = sum(sensitivity)
                individual.costs_signed[-2] = sum(sensitivity)
            else:
                individual.costs.append(sum(sensitivity))
                individual.costs_signed.insert(-1, sum(sensitivity))


class Generator(Operator):
    def __init__(self, parameters=None, features=dict()):
        super().__init__()
        self.parameters = parameters
        self.features = features

    @abstractmethod
    def generate(self):
        pass


class CustomGenerator(Generator):
    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)
        self.vectors = []

    def init(self, vectors):
        self.vectors = vectors

    def generate(self):
        individuals = []
        for vector in self.vectors:
            individuals.append(Individual(vector, self.features))
        return individuals


class RandomGenerator(Generator):
    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)
        self.number = 0

    def init(self, number):
        self.number = number

    def generate(self):
        individuals = []
        for i in range(self.number):
            vector = VectorAndNumbers.gen_vector(self.parameters)
            individuals.append(Individual(vector, self.features))
        return individuals


class FullFactorGenerator(Generator):
    """
    Create a general full-factorial design
    Number of experiments (2 ** len(parameters) - without center, 3 ** len(parameters - with center)
    """

    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)
        self.center = False

    def init(self, center):
        self.center = center

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]

            if self.center:
                dict_vars[name] = [l_b, (l_b + u_b) / 2.0, u_b]
            else:
                dict_vars[name] = [l_b, u_b]

        df = build_full_fact(dict_vars)

        individuals = []
        for vector in df:
            individuals.append(Individual(vector, self.features))
        return individuals


class PlackettBurmanGenerator(Generator):
    """
    Create a general full-factorial design
    Number of experiments (2 ** len(parameters) - without center, 3 ** len(parameters - with center)
    """

    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_plackett_burman(dict_vars)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(Individual(vector, self.features))
        return individuals


class BoxBehnkenGenerator(Generator):
    """
    Create a general full-factorial design
    # 3 params = 13 experiments
    # 4 params = 25 experiments
    # 5 params = 41 experiments
    # 6 params = 49 experiments
    # 7 params = 57 experiments
    # 8 params = 113 experiments
    https://en.wikipedia.org/wiki/Box%E2%80%93Behnken_design
    """

    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter['name']
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_box_behnken(dict_vars)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(Individual(vector, self.features))
        return individuals


class LHSGenerator(Generator):
    """
    Builds a Latin Hypercube design dataframe from a dictionary of factor/level ranges.
    """

    def __init__(self, parameters=None, features=dict()):
        super().__init__(parameters, features)
        self.number = 0

    def init(self, number):
        self.number = number

    def generate(self):
        dict_vars = {}
        for parameter in self.parameters:
            name = parameter["name"]
            l_b = parameter['bounds'][0]
            u_b = parameter['bounds'][1]
            dict_vars[name] = [l_b, u_b]

        df = build_lhs(dict_vars, num_samples=self.number)
        # print(df)

        individuals = []
        for vector in df:
            individuals.append(Individual(vector, self.features))
        return individuals


class Mutator(Operator):
    def __init__(self, parameters, probability):
        super().__init__()
        self.parameters = parameters
        self.probability = probability

    @abstractmethod
    def mutate(self, p, current_iteration=0):
        pass


class SimpleMutator(Mutator):
    def __init__(self, parameters, probability):
        super().__init__(parameters, probability)

    def mutate(self, p, current_iteration=0):
        """ uniform random mutation """
        mutation_space = 0.1
        vector = []

        for i, parameter in enumerate(self.parameters):
            if random.uniform(0, 1) < self.probability:
                l_b = parameter['bounds'][0]
                u_b = parameter['bounds'][1]

                para_range = mutation_space * (u_b - l_b)
                mutation = random.uniform(-para_range, para_range)
                # vector.append(self.clip(p.vector[i] + mutation, l_b, u_b))
                vector.append(p.vector[i] + mutation)
            else:
                vector.append(p.vector[i])

        p_new = Individual(vector, p.features)
        return p_new


class PmMutator(Mutator):
    """
    PmMutation -- for nsga2 and epsMoEA

    This operator can handle real, integer and boolean optimization parameters.
    The class contains two-kind of operators as the original Deb's implementation .
    """

    def __init__(self, parameters, probability, distribution_index=20):
        super().__init__(parameters, probability)
        self.distribution_index = distribution_index

    def mutate(self, parent, current_iteration=0):
        vector = []

        for i, parameter in enumerate(self.parameters):
            if random.uniform(0, 1) < self.probability:
                l_b = parameter['bounds'][0]
                u_b = parameter['bounds'][1]
                vector.append(self.pm_mutation(parent.vector[i], l_b, u_b))
            else:
                vector.append(parent.vector[i])

        return Individual(vector, parent.features)

    def pm_mutation(self, x, lb, ub):
        """
        Polynomial mutation for float and integer parameters.

        :param x: represents one parameter of the problem
        :param lb: lower bound
        :param ub: upper bound
        :return:
        """
        rnd = random.uniform(0, 1)
        dx = ub - lb
        delta1 = (x - lb) / dx
        delta2 = (ub - x) / dx
        mut_pow = 1.0 / (self.distribution_index + 1.0)
        if rnd < 0.5:
            xy = 1.0 - delta1
            val = 2.0 * rnd + (1.0 - 2.0 * rnd) * (pow(xy, self.distribution_index + 1.0))
            deltaq = pow(val, mut_pow) - 1.0
        else:
            xy = 1.0 - delta2
            val = 2.0 * (1.0 - rnd) + 2.0 * (rnd - 0.5) * (pow(xy, self.distribution_index + 1.0))
            deltaq = 1.0 - pow(val, mut_pow)

        x = x + deltaq * dx
        x = self.clip(x, lb, ub)

        return x

    def bitflip(self, x: list):

        for j in range(1, len(x)):
            if random.uniform(0.0, 1.0) <= self.probability:
                x[j] = not x[j]

        return x


class UniformMutator(Mutator):
    """
    Uniform mutator -- for omopso .
    """

    def __init__(self, parameters, probability, perturbation=0.5):
        super().__init__(parameters, probability)
        self.perturbation = perturbation

    def mutate(self, parent, current_iteration=0):
        vector = []

        for i, parameter in enumerate(self.parameters):
            if random.uniform(0, 1) < self.probability:
                l_b = parameter['bounds'][0]
                u_b = parameter['bounds'][1]
                vector.append(self.uniform_mutation(parent.vector[i], l_b, u_b))
            else:
                vector.append(parent.vector[i])

        return Individual(vector, parent.features)

    def uniform_mutation(self, x, lb, ub):

        x = x + (random.random() - 0.5) * self.perturbation
        x = self.clip(x, lb, ub)

        return x


class NonUniformMutation(Mutator):

    def __init__(self, parameters, probability, max_iterations, perturbation=0.5, ):
        super().__init__(parameters, probability)
        self.perturbation = perturbation
        self.max_iterations = max_iterations

    def __delta(self, y: float, b_mutation_parameter: float, current_iteration):
        return (y * (1.0 - pow(random.random(),
                               pow((1.0 - 1.0 * current_iteration / self.max_iterations), b_mutation_parameter))))

    def mutate(self, parent, current_iteration=0):
        vector = []

        for i, parameter in enumerate(self.parameters):
            if random.uniform(0, 1) < self.probability:
                l_b = parameter['bounds'][0]
                u_b = parameter['bounds'][1]
                vector.append(self.non_uniform_mutation(parent.vector[i], l_b, u_b, current_iteration))
            else:
                vector.append(parent.vector[i])

        return Individual(vector, parent.features)

    def non_uniform_mutation(self, x, lb, ub, current_iteration):

        rand = random.random()

        if rand <= 0.5:
            x = self.__delta(ub - x, self.perturbation, current_iteration)
        else:
            x = self.__delta(lb - x, self.perturbation, current_iteration)

        if isinstance(x, complex):
            print(x)
        x = self.clip(x, lb, ub)

        return x


# class SwarmStep(Mutator):
#     """
#     This swarm mutator operator is made for the original PSO algorithm, which defined by Kennedy and Eberhart in 1995
#
#     PSO shares many similarities with evolutionary computation. Both algorithms start with a group of a randomly
#     generated population. Both update the population iteratively and search for the optimum with stochastic techniques.
#     The main difference between them is in the information sharing mechanism. In EA, only the individuals of current
#     generation share information with each other, and any individual has a chance to give out information to others.
#     In PSO, actually not the current individuals share information with each other, but the individuals of previous
#     generation (the optimal particles) give out information to the current ones. In other words, the information sharing
#     is one-way in PSO.
#     """
#
#     def __init__(self, parameters, probability=1):
#         super().__init__(parameters, probability)
#         self.w = 0.1  # constant inertia weight (how much to weigh the previous velocity)
#         self.c1 = 2.  # cognitive constant
#         self.c2 = 1.  # social constant
#         self.best_individual = None
#
#     def evaluate_best_individual(self, individual):
#         """ Determines the best individual in the swarm """
#         dominates = True
#         individual.features['best_costs'] = [0] * len(individual.costs)
#         for i in range(len(individual.features['best_costs'])):
#             if individual.costs[i] > individual.features['best_costs'][i]:
#                 dominates = False
#
#         # check to see if the current position is an individual best
#         if dominates:
#             individual.features['best_vector'] = individual.vector
#             individual.features['best_costs'] = individual.costs
#
#     # update new particle velocity
#     def update_velocity(self, individual):
#         individual.features['velocity'] = [0] * len(individual.vector)
#         individual.features['best_vector'] = [0] * len(individual.vector)
#         for i in range(0, len(individual.vector)):
#             r1 = 0.1 * random.random()
#             r2 = 0.1 * random.random()
#
#             vel_cognitive = self.c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
#             vel_social = self.c2 * r2 * (self.best_individual.vector[i] - individual.vector[i])
#             individual.features['velocity'][i] = self.w * individual.features['velocity'][
#                 i] + vel_cognitive + vel_social
#
#     # update the particle position based off new velocity updates
#     def update_position(self, individual):
#
#         for parameter, i in zip(self.parameters, range(len(individual.vector))):
#             individual.vector[i] = individual.vector[i] + individual.features['velocity'][i]
#
#             # adjust maximum position if necessary
#             if individual.vector[i] > parameter['bounds'][1]:
#                 individual.vector[i] = parameter['bounds'][1]
#
#             # adjust minimum position if necessary
#             if individual.vector[i] < parameter['bounds'][0]:
#                 individual.vector[i] = parameter['bounds'][0]
#
#     def update(self, best_individual):
#         self.best_individual = best_individual
#
#     def mutate(self, p):
#         self.update_velocity(p)
#         self.update_position(p)
#         return p
#

# class SwarmStepTVIW(SwarmStep):
#     """
#     This is an improvement of the original PSO algorithm with Time Varying Inertia Weight operators.
#
#     Empirical study of particle swarm optimization,” in Proc. IEEE Int. Congr. Evolutionary Computation, vol. 3,
#     1999, pp. 101–106.
#
#     Shi and Eberhart have observed that the optimal solution can be improved by varying the inertia weight value from
#     0.9 (at the beginning of the search) to 0.4 (at the end of the search) for most problems. This modification to the
#     original PSO concept has been considered as the basis for two novel strategies introduced in this paper. Hereafter,
#     in this paper, this version of PSO is referred to as time-varying inertia weight factor method
#
#     Contras:
#     -------
#     - PSO-TVIW concept is not very effective for tracking dynamic systems
#
#     - its ability to fine tune the optimum solution is comparatively weak, mainly due
#       to the lack of diversity at the end of the search
#
#     R. C. Eberhart and Y. Shi, “Tracking and optimizing dynamic systems with particle swarms,” in Proc. IEEE Congr.
#     Evolutionary Computation 2001, Seoul, Korea, 2001, pp. 94–97
#     """
#
#     def __init__(self, parameters, probability=1, nr_maxgen=100):
#         super().__init__(parameters, probability)
#         self.w1 = 0.9  # inertia weight is calculated from w1 and w2
#         self.w2 = 0.4
#         self.c1 = 2.  # cognitive constant
#         self.c2 = 1.  # social constant
#         self.best_individual = None
#
#         # new parameters
#         self.max_nr_generations = nr_maxgen
#         self.current_iter = 0.
#
#     # update new particle velocity
#     def update_velocity(self, individual):
#         """
#         :param nr_generations: total number of generations, during the calculation, MAXITER
#         :param iteration_nr: actual generation
#         """
#
#         individual.features['velocity'] = [0] * len(individual.vector)
#         individual.features['best_vector'] = [0] * len(individual.vector)
#         for i in range(0, len(individual.vector)):
#             r1 = 0.1 * random.random()
#             r2 = 0.1 * random.random()
#
#             # (w1-w2)*(MAX_ITER-iter)/MAX_ITER
#             w = (self.w1 - self.w2) * (self.max_nr_generations - self.current_iter) / self.max_nr_generations + self.w2
#
#             vel_cognitive = self.c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
#             vel_social = self.c2 * r2 * (self.best_individual.vector[i] - individual.vector[i])
#             individual.features['velocity'][i] = w * individual.features['velocity'][i] + vel_cognitive + vel_social
#
#             self.current_iter += 1.
#

# class SwarmStepRandIW(SwarmStep):
#     """
#     In this variation, the inertia weght is changing randomly,the mean value of the inertia weight is 0.75.
#     This modification was inspired by Clerc’s constriction factor concept,  in which the inertia weight is
#     kept constant at 0.729 and both acceleration coefficients are kept constant at 1.494.
#     Therefore, when random inertia weight factor method is used the acceleration coefficients are kept constant at 1.494.
#
#     Contras:
#     -------
#     """
#
#     def __init__(self, parameters, probability=1):
#         super().__init__(parameters, probability)
#         self.w = 0.5  # inertia weight -> changed randomly
#         self.c1 = 2.  # cognitive constant
#         self.c2 = 1.  # social constant
#         self.best_individual = None
#
#     # update new particle velocity
#     def update_velocity(self, individual):
#         """
#         :param nr_generations: total number of generations, during the calculation, MAXITER
#         :param iteration_nr: actual generation
#         """
#         for i in range(0, len(individual.vector)):
#             r1 = 0.1 * random.random()
#             r2 = 0.1 * random.random()
#
#             # (w1-w2)*(MAX_ITER-iter)/MAX_ITER
#             w = self.w * random.random() / 2.
#
#             vel_cognitive = self.c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
#             vel_social = self.c2 * r2 * (self.best_individual.vector[i] - individual.vector[i])
#             individual.velocity_i[i] = w * individual.velocity_i[i] + vel_cognitive + vel_social
#

# class FireflyStep(SwarmStep):
#     """
#     Firefly algorithm is a modification of the original pso algorithms. The idea is that it mimics the behaviour of
#     the fireflies, which uses specfic light combinations for hunting and dating. This algorithm mimics the dating
#     behaviour of these bugs. The algorithm is originally published by [1]
#
#     The brightest individual attracts the darkest ones, this starts to go to that direction, if there is not an
#     existing brighter solutions, the algorithm randomly steps one into another direction.
#
#     This operator makes the following
#     ---------------------------------
#
#     - calculates the distance between two points, because its correlates with that value.
#     - the brightness of the other point
#     - the mutator constant, which contains a damping factor [2]?
#
#     The k+1 th position of the j(th) individual is calculated by the following formula
#     ---------------------------------------------------------------------------
#
#         b0=2;               # Attraction Coefficient Base Value
#         a=0.2;              # Mutation Coefficient
#         ad=0.98;            # Mutation Coefficient Damping Ratio -- decreases the initial value of a after each
#                               iteration step
#         gamma = 1.          # light absorbtion coefficient
#
#         x(j,k+1) = x(j,k) + b0*exp(-gamma*r^2) + sum_{i<j}[x(j,k) - x(i,k)] + a*(rand[0,1] - 0.5)
#
#     [1] Yang, Xin-She. "Firefly algorithms for multimodal optimization." International symposium on stochastic algorithms.
#         Springer, Berlin, Heidelberg, 2009.
#     [2] firefly algorithm implementation from www.Yarpiz.com
#
#     [3] https://nl.mathworks.com/matlabcentral/fileexchange/29693-firefly-algorithm
#
#      Similarly, alpha should also be linked with scales, the steps should not too large or too small, often steps
#      are about 1/10 to 1/100 of the domain size. In addition, alpha should be reduced gradually using
#      alpha=alpha_0 delta^t during eteration t.  Typically, delta=0.9 to 0.99 will be a good choice.
#     """
#
#     def __init__(self, parameters, probability=0.05):
#         super().__init__(parameters, probability)
#
#         self.beta = 2.0
#         self.alpha = 0.2
#         self.ad = 0.98
#         self.gamma = 1.
#
#         self.best_individual = None
#
#     def dominate(self, current, other):
#         """True if other dominates over current """
#         dominates = True
#
#         for i in range(len(current.costs)):
#             if other.costs_signed[i] > current.costs_signed[i]:
#                 dominates = False
#
#         return dominates
#
#     def update_coefficient_a(self):
#         """ Updates the mutation coefficient with the damping factor, after each iteration step """
#         self.alpha *= self.ad
#         return
#
#     # update new particle velocity
#     def update_velocity_ij(self, current, other):
#         """
#         This algorithm has a two-layered hierarchy, because every individual calculates an approximative next position
#         from the light intensity between two selected points.
#         """
#         r2 = 0.  # euclidean distance
#
#         for i, param in enumerate(self.parameters):
#             lb = param['bounds'][0]
#             ub = param['bounds'][1]
#
#             # elementary distance of the particle
#             e = self.probability * (ub - lb)
#             # distance between the two individuals
#             r2 += current.vector[i] ** 2. + other.vector[i] ** 2.
#             vel_attraction = self.beta * exp(-self.gamma * r2 ** 2.)
#
#             v_rd = self.alpha * (random.random() - 0.5) * e
#             current.velocity_i[i] = vel_attraction + v_rd
#
#             return
#
#     def mutate_ij(self, p, q):
#         """
#         For firefly algorithm, because of the two-layered hierarchy.
#
#         :param p:
#         :param q:
#         :return:
#         """
#         self.update_velocity_ij(p, q)
#         self.update_position(p)
#         return
#
#
# class SwarmStepTVAC(SwarmStep):
#     """
#     Time-varying acceleration coefficients as a new parameter automation strategy for the PSO concept.
#
#     An improved optimum solution for most of the benchmarks was observed when changing c1 from 2.5 to 0.5
#     and changing c2 from 0.5 to 2.5, over the full range of the search.
#     Therefore, these values are used for the rest of the work. With this modification, a significant improvement of
#     the optimum value and the rate of convergence were observed, particularly for unimodal functions, compared with
#     the PSO-TVIW. However, it has been observed that the performance of the PSO-TVAC method is similar or poor
#     for multimodal functions. In contrast, compared with the PSO-RANDIW method an improved performance has been
#     observed with the PSO-TVAC for multimodal functions.
#     However, for unimodal functions, the PSO-RANDIW method showed significantly quick convergence to a good solution
#     compared with the PSO-TVAC method. The results are presented and discussed in Section V.
#
#     Pros: improved convergence rate in case of multi-modal functions
#
#     Cons: significantly slower convergence rate than PSO-RandIW for unimodal functions
#     """
#
#     def __init__(self, parameters, probability=1):
#         super().__init__(parameters, probability)
#         self.w = 0.9  # inertia weight
#         self.c1i = 0.5  # cognitive constant initial value
#         self.c1f = 2.5  # cognitive constant final value
#         self.c2i = 2.5  # social constant initial value
#         self.c2f = 0.5  # social constant final value
#         self.best_individual = None
#
#     # update new particle velocity
#     def update_velocity(self, individual, nr_generations, iteration_nr):
#         """
#         :param nr_generations: total number of generations, during the calculation, MAXITER
#         :param iteration_nr: actual generation
#         """
#         for i in range(0, len(individual.vector)):
#             r1 = 0.1 * random.random()
#             r2 = 0.1 * random.random()
#
#             # (c1f-c1i)*(MAX_ITER-iter)/MAX_ITER +
#             c1 = (self.c1f - self.c1i) * (nr_generations - iteration_nr) / nr_generations + self.c1i
#             c2 = (self.c2f - self.c2i) * (nr_generations - iteration_nr) / nr_generations + self.c2i
#
#             vel_cognitive = c1 * r1 * (individual.features['best_vector'][i] - individual.vector[i])
#             vel_social = c2 * r2 * (self.best_individual.vector[i] - individual.vector[i])
#             individual.velocity_i[i] = self.w * individual.velocity_i[i] + vel_cognitive + vel_social
#

class Dominance(ABC):
    def __init__(self):
        pass

    def compare(self, p, q):
        raise NotImplementedError("method not implemented")


class EpsilonDominance(Dominance):
    """
    Epsilon dominance comparator divides the number
    """

    def __init__(self, epsilons):
        super(EpsilonDominance, self).__init__()

        if hasattr(epsilons, "__getitem__"):
            self.epsilons = epsilons
        else:
            self.epsilons = [epsilons]

    def compare(self, p: list, q: list):
        """
        Eps Dominance comparator, the implementation based on Platypus.
        This operator works similarly like the normal, Dominance operator. However, it divides the current value with
        the epsilons and selects which dominates the other. If they are equal it selects those one, which is closer to
        the boundary.

        :param p: solution p
        :param q: solution q
        :return:
        """

        # first check constraint violation, the last item is the feasibility, which is a real number if its zero,
        # it means that the solution is feasible
        if p[-1] != q[-1]:
            if p[-1] == 0:
                return 1  # p dominates
            elif q[-1] == 0:
                return 2  # q is dominates, because it has smaller degree in constraint violation
            elif abs(p[-1]) < abs(q[-1]):
                return 1  # p is dominates
            elif abs(q[-1]) < abs(p[-1]):
                return 2  # q is dominates

        # then use epsilon dominance on the objectives
        dominate_p = False
        dominate_q = False

        for i, (p_costs, q_costs) in enumerate(zip(p[:-1], q[:-1])):

            epsilon = float(self.epsilons[i % len(self.epsilons)])

            p_eps = math.floor(p_costs / epsilon)
            q_eps = math.floor(q_costs / epsilon)

            if p_eps > q_eps:
                dominate_q = True
                if dominate_p:
                    return 0
            elif q_eps > p_eps:
                dominate_p = True
                if dominate_q:
                    return 0

        if not dominate_p and not dominate_q:
            dist1 = 0.0
            dist2 = 0.0

            for i, (p_costs, q_costs) in enumerate(zip(p[:-1], q[:-1])):
                epsilon = float(self.epsilons[i % len(self.epsilons)])

                i1 = math.floor(p_costs / epsilon)
                i2 = math.floor(q_costs / epsilon)

                dist1 += math.pow(p_costs - i1 * epsilon, 2.0)
                dist2 += math.pow(q_costs - i2 * epsilon, 2.0)

            if dist1 < dist2:
                return 1
            else:
                return 2
        elif dominate_p:
            return 1
        else:
            return 2

    def same_box(self, p, q):
        # if not dominate1 and not dominate2:
        if self.compare(p, q) == 0:
            return True
        else:
            return False


class ParetoDominance(Dominance):
    """Pareto dominance with constraints.

    If either solution violates constraints, then the solution with a smaller
    constraint violation is preferred. If both solutions are feasible, then
    Pareto dominance is used to select the preferred solution.
    """

    def __init__(self, epsilons=None):
        super(ParetoDominance, self).__init__()
        # self.signs = signs deprecated

    def compare(self, p: list, q: list):
        """
        Here, p and q are tuples, which contains the (feasibility index, cost vector)
        """
        # assert len(p) == len(q)

        # first check constraint violation, the last item is the feasibility, which is a real number if its zero,
        # it means that the solution is feasible
        if p[-1] != q[-1]:
            if p[-1] == 0:
                return 1  # p dominates
            elif q[-1] == 0:
                return 2  # q is dominates, because it has smaller degree in constraint violation
            elif abs(p[-1]) < abs(q[-1]):
                return 1  # p is dominates
            elif abs(q[-1]) < abs(p[-1]):
                return 2  # q is dominates

        dominate_p = False
        dominate_q = False

        for (p_costs, q_costs) in zip(p[:-1], q[:-1]):
            if p_costs > q_costs:
                dominate_q = True
                if dominate_p:
                    return 0
            elif q_costs > p_costs:
                dominate_p = True
                if dominate_q:
                    return 0

        if dominate_q == dominate_p:
            return 0
        elif dominate_p:
            return 1
        else:
            return 2


class Selector(Operator):

    def __init__(self, parameters, sign=None, part_num=2, dominance=ParetoDominance):
        """

        :param parameters:
        :param sign: one value from now, because its goal is to tell the direction of the optimum min or max
        :param part_num:
        :param dominance:
        """
        super().__init__()
        self.parameters = parameters
        self.comparator = dominance()  # ParetoDominance()
        self.signs = sign

    @abstractmethod
    def select(self, population):
        pass

    def pop_acceptance(self, individuals, individual):
        """
        This function differs from add() function that it preserves the length of the _content in the archive class.
        :return:
        """

        dominates = []
        dominated = False

        for i in range(len(individuals)):
            flag = self.dominance.compare(individual.costs_signed, individuals[i].costs_signed)

            if flag == 1:
                dominates.append(i)
            elif flag == 2:
                dominated = True

        if len(dominates) > 0:
            del individuals[random.choice(dominates)]
            individuals.append(individual)

        elif not dominated:
            individuals.remove(random.choice(individuals))
            individuals.append(individual)

        return

    def fast_nondominated_sorting(self, population):

        pareto_front = [[]]
        front_number = 1

        # reset elements
        for p in population:
            p.features['domination_counter'] = 0
            p.features['front_number'] = None
            p.features['dominate'] = set()

        for i, p in enumerate(population):
            for j in range(i + 1, len(population)):
                q = population[j]
                dom = self.comparator.compare(p.costs_signed, q.costs_signed)
                if dom == 1:
                    p.features['dominate'].add(q)
                    q.features['domination_counter'] += 1
                elif dom == 2:
                    p.features['domination_counter'] += 1
                    q.features['dominate'].add(p)

            # selects the pareto values
            if p.features['domination_counter'] == 0:
                p.features['front_number'] = front_number
                pareto_front[front_number - 1].append(p)

        while len(pareto_front[front_number - 1]) > 0:
            front_number += 1
            pareto_front.append([])
            for p in pareto_front[front_number - 2]:
                for q in p.features['dominate']:
                    q.features['domination_counter'] -= 1
                    if q.features['domination_counter'] == 0 and q.features['front_number'] is None:
                        q.features['front_number'] = front_number
                        pareto_front[front_number - 1].append(q)

        if len(pareto_front[front_number - 1]) == 0:
            pareto_front.pop()
        for sub_front in pareto_front:
            crowding_distance(sub_front)

        return


def crowding_distance(front):
    """
    Crowding distance calculates the solution density on a front, a subset of the population.
    :param front: list of individuals
                  which is a subset of the total population
    :return:
    """
    n = len(front)

    if n == 0:
        return
    elif n == 1:
        front[0].features['crowding_distance'] = math.inf
        return
    elif n == 2:
        front[0].features['crowding_distance'] = math.inf
        front[1].features['crowding_distance'] = math.inf
        return

    for i in range(len(front)):
        front[i].features['crowding_distance'] = 0.0

    for dim in range(len(front[0].costs_signed[:-1])):

        front.sort(key=lambda x: x.costs_signed[dim])
        # self.sort_by_coordinate(population, dim)

        front[0].features['crowding_distance'] = math.inf
        front[-1].features['crowding_distance'] = math.inf
        max_distance = front[-1].costs_signed[dim] - front[0].costs_signed[dim]
        for i in range(1, n - 1):
            distance = front[i + 1].costs_signed[dim] - front[i - 1].costs_signed[dim]
            if max_distance > 0.0:
                front[i].features['crowding_distance'] += distance / max_distance
    return


def nondominated_truncate(population, size):
    """Truncates a population to the given size, using non-dominated sorting.
    The resulting population is filled with the first N-1 fronts.
    The Nth front is too large and must be split using crowding distance.

    Parameters
    ----------
    :population : iterable
        The collection of solutions that have been non-domination sorted
    :size:
        The size of the truncated result
    """

    # calculating the crowding distance on the different fronts
    population = list(set(population))
    result = sorted(population, key=functools.cmp_to_key(nondominated_cmp))
    return result[:size]


def nondominated_cmp(p, q):
    """
    From the 'front_number' the smaller value is favourized.
    From the 'crowding_distance' the higher values are favoured.
    :param p: Individual
    :param q: Individual
    :return: -1 if x dominates
              0 if they are equals
              1 if y dominates
    """
    if p.features['front_number'] == q.features['front_number']:
        if -p.features['crowding_distance'] < -q.features['crowding_distance']:
            return -1
        elif -p.features['crowding_distance'] > -q.features['crowding_distance']:
            return 1
        else:
            return 0
    else:
        if p.features['front_number'] < q.features['front_number']:
            return -1
        elif p.features['front_number'] > q.features['front_number']:
            return 1
        else:
            return 0


class DummySelector(Selector):
    def __init__(self, parameters):
        super().__init__(parameters)

    def select(self, individuals):
        return individuals


class CopySelector(Selector):
    def __init__(self, parameters):
        super().__init__(parameters)

    def select(self, individuals):
        selection = []
        for individual in individuals:
            candidate = Individual(individual.vector)
            candidate.costs = deepcopy(individual.costs)
            candidate.features = deepcopy(individual.features)
            candidate.population_id = -1
            selection.append(candidate)
        return selection


class TournamentSelector(Selector):
    def __init__(self, parameters, dominance=ParetoDominance, epsilons=None):
        super().__init__(parameters)
        self.dominance = dominance(epsilons=epsilons)

    def select(self, individuals):
        """
        Binary tournament selection:

        2 individuals are selected from the given population (randomly), then the function gaves
        back the dominant (or random).
        """

        if len(individuals) == 1:
            selected = individuals[0]
        else:
            # Sampling two individuals without a replacement
            candidates = random.sample(individuals, 2)

            # they should have this at this stage
            if candidates[0].features['front_number'] < candidates[1].features['front_number']:
                return candidates[0]
            elif candidates[1].features['front_number'] < candidates[0].features['front_number']:
                return candidates[1]

            flag = self.dominance.compare(candidates[0].costs_signed, candidates[1].costs_signed)

            if flag == 1:
                selected = candidates[0]
            elif flag == 2:
                selected = candidates[1]
            else:
                selected = random.choice(candidates)

        return selected


class FireflyStep(Operator):
    """
      This operator makes the following
      ---------------------------------

      - calculates the distance between two points, because its correlates with that value.
      - the brightness of the other point
      - the mutator constant, which contains a damping factor [2]?

      The k+1 th position of the j(th) individual is calculated by the following formula
      ---------------------------------------------------------------------------

          L = ub-lb
          b=1;               # Attraction Coefficient Base Value, if beta = 0, it becamos a random walk
          a=0.01*L;              # Mutation Coefficient
          ad=0.9;            # Mutation Coefficient Damping Ratio -- decreases the initial value of a after each, as same as in simulated annealing
                                iteration step
          gamma = 0.5/L^2          # light absorbtion coefficient


          x(j,k+1) = x(j,k) + b0*exp(-gamma*r^2) + sum_{i<j}[x(j,k) - x(i,k)] + a*(rand[0,1] - 0.5)

      [1] Yang, Xin-She. "Firefly algorithms for multimodal optimization." International symposium on stochastic algorithms.
          Springer, Berlin, Heidelberg, 2009.

      [2]: Yang, X. S. (2013). Multiobjective firefly algorithm for continuous optimization.
              Engineering with computers, 29(2), 175-184.
    """

    def __init__(self, parameters, alpha = 0.01, beta = 1.0, gamma = 0.5, damping = 0.9, dominance=ParetoDominance()):
        super().__init__()
        self.parameters = parameters
        self.dominance = dominance
        self.b = beta
        self.a = alpha
        self.gamma = gamma
        self.damping_factor = damping

    def attraction_step(self, current, other, iteration_nr):
        # refresh the position of the offspring if its more attractive
        # if it's dominant returns without changing its position
        if self.dominance.compare(current.costs_signed, other.costs_signed) == 1:
            return
        else:
            for i, param in enumerate(self.parameters):
                lb = param['bounds'][0]
                ub = param['bounds'][1]

                L = (ub-lb)
                Gamma = L*self.gamma
                alpha = self.a*L*0.9**iteration_nr

                # elementary distance of the particle, random walk
                e = random.random() * (ub - lb)
                # distance between the two individuals
                r2 = current.vector[i] ** 2. + other.vector[i] ** 2.
                vel_attraction = self.b * exp(-Gamma * r2 ** 2.) * (other.vector[i] - current.vector[i])

                v_rd = alpha * (random.random() - 0.5) * e
                # position i+1
                current.vector[i] += vel_attraction + v_rd
                current.vector[i] = self.clip(current.vector[i], lb, ub)
        return


class Crossover(Operator):

    def __init__(self, parameters, probability):
        super().__init__()
        self.parameters = parameters
        self.probability = probability

    @abstractmethod
    def cross(self, p1, p2):
        pass


class SimpleCrossover(Crossover):
    def __init__(self, parameters, probability):
        super().__init__(parameters, probability)

    def cross(self, p1, p2):
        parent_a = p1.vector.copy()
        parent_b = p2.vector.copy()

        if random.uniform(0.0, 1.0) <= self.probability:
            parameter1 = []
            parameter2 = []
            linear_range = 2

            alpha = random.uniform(0, linear_range)

            for i, param in enumerate(self.parameters):
                l_b = param['bounds'][0]
                u_b = param['bounds'][1]

                parameter1.append(self.clip(alpha * p1.vector[i] + (1 - alpha) * p2.vector[i], l_b, u_b))
                parameter2.append(self.clip((1 - alpha) * p1.vector[i] + alpha * p2.vector[i], l_b, u_b))

            parent_a = parameter1
            parent_b = parameter2

        offspring_a = Individual(parent_a, p1.features)
        offspring_b = Individual(parent_b, p2.features)

        return offspring_a, offspring_b


class SimulatedBinaryCrossover(Crossover):

    def __init__(self, parameters, probability, distribution_index=15):
        super().__init__(parameters, probability)
        self.distribution_index = distribution_index
        if distribution_index < 0:
            raise Exception('The distribution index have to be positive.')
        if probability > 1.:
            raise Exception('The probability should be selected from [0,1].')
        elif probability < 0.:
            raise Exception('The probability should be selected from [0,1].')

    def cross(self, p1, p2):
        """
        Create an offspring using simulated binary crossover.
        :return:  Individual, Individual

        a list with 2 offsprings each with the genotype of an  offspring after recombination and mutation.
        """

        x1 = deepcopy(p1.vector)
        x2 = deepcopy(p2.vector)

        if random.random() <= self.probability:
            for i, param in enumerate(self.parameters):

                if random.random() <= 0.5:

                    if abs(x2[i] - x1[i]) > EPSILON:
                        if x2[i] > x1[i]:
                            y1, y2 = x1[i], x2[i]
                        else:
                            y1, y2 = x2[i], x1[i]

                        lb, ub = param['bounds'][0], param['bounds'][1]

                        # calculates c1
                        beta = 1.0 + (2.0 * (y1 - lb) / (y2 - y1))
                        alpha = 2.0 - pow(beta, -(self.distribution_index + 1.0))

                        rand = random.random()
                        if rand <= (1.0 / alpha):
                            betaq = pow(rand * alpha, (1.0 / (self.distribution_index + 1.0)))
                        else:
                            betaq = pow(1.0 / (2.0 - rand * alpha), 1.0 / (self.distribution_index + 1.0))

                        c1 = 0.5 * (y1 + y2 - betaq * (y2 - y1))

                        # calculates c2
                        beta = 1.0 + (2.0 * (ub - y2) / (y2 - y1))
                        alpha = 2.0 - pow(beta, -(self.distribution_index + 1.0))

                        if rand <= (1.0 / alpha):
                            betaq = pow((rand * alpha), (1.0 / (self.distribution_index + 1.0)))
                        else:
                            betaq = pow(1.0 / (2.0 - rand * alpha), 1.0 / (self.distribution_index + 1.0))

                        c2 = 0.5 * (y1 + y2 + betaq * (y2 - y1))

                        # check the boundaries
                        c1 = self.clip(c1, lb, ub)
                        c2 = self.clip(c2, lb, ub)

                        if random.random() <= 0.5:
                            x1[i], x2[i] = c2, c1
                        else:
                            x1[i], x2[i] = c1, c2

        return Individual(x1, p1.features), Individual(x2, p2.features)



import textwrap
import re
import tempfile
import os
import subprocess
import datetime
import time
import ntpath
import pathlib
from string import Template

import rpyc
from rpyc.core.async_ import AsyncResultTimeout
from rpyc.utils.classic import upload_file, download_file

from abc import ABCMeta, abstractmethod
from .utils import ConfigDictionary
from shutil import copyfile

from .config import config


# parse ip address
def parse_address(address):
    regex_match = re.compile(r'<(\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}):.*').match(address)
    if regex_match is not None:
        return regex_match.group(1)
    return ""


class Executor(metaclass=ABCMeta):

    """
    Function is a class representing objective or cost function for
    optimization problems.
    """

    def __init__(self, problem):
        self.problem = problem

        self.options = ConfigDictionary()
        # parse method
        self.parse_results = None

    @abstractmethod
    def eval(self, x):
        # set default parse method from problem (can be overridden)
        if self.parse_results is None and "parse_results" in dir(self.problem):
            self.parse_results = self.problem.parse_results

    @staticmethod
    def _join_parameters_names(parameters, sep=","):
        param_names_string = ""

        for parameter in parameters:
            param_names_string += str(parameter['name']) + sep

        # remove last sep
        if (len(parameters)) > len(sep):
            param_names_string = param_names_string[:-len(sep)]

        return param_names_string

    @staticmethod
    def _join_parameters_values(x, sep=","):
        param_values_string = ""
        for number in x:
            param_values_string += str(number) + sep

        # remove last sep
        if (len(x)) > len(sep):
            param_values_string = param_values_string[:-len(sep)]

        return param_values_string

    @staticmethod
    def _remove_dir(local_dir):
        files = os.listdir(path=local_dir)
        for f in files:
            filepath = os.path.join(local_dir, f)
            os.remove(filepath)

        os.rmdir(local_dir)


class LocalComsolExecutor(Executor):

    comsol_command = Template("comsol batch -inputfile $input_file -nosave -pname $param_names -plist $param_values")

    def __init__(self, problem, problem_file, input_files=None, output_files=None):
        super().__init__(problem)
        self.model_file = ntpath.basename(problem_file)
        copyfile(problem_file, self.problem.working_dir + self.model_file)

        self.output_files = output_files
        self.input_files = input_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        cmd_string = self.comsol_command.substitute(input_file=self.model_file,
                                                    param_names=param_names_string,
                                                    param_values=param_values_string)

        try:
            # run command
            current_path = os.getcwd()
            os.chdir(self.problem.working_dir)

            out = subprocess.run(cmd_string, shell=True)
            os.chdir(current_path)

            if out.returncode != 0:
                err = "Unknown error"
                if out.stderr is not None:
                    err = "Cannot run COMSOL Multiphysics.\n\n {}".format(out.stderr)

                self.problem.logger.error(err)
                raise RuntimeError(err)

            output_files = []
            for file in self.output_files:
                output_files.append(self.problem.working_dir + file)
            result = self.parse_results(output_files, individual)

            return result
        except Exception as e:
            err = "Cannot run COMSOL Multiphysics.\n\n {}".format(e)
            self.problem.logger.error(err)
            raise RuntimeError(err)


class RemoteExecutor(Executor):
    """
        Allows distributing of calculation of objective functions.
        """

    def __init__(self, problem, command, files_to_server, files_from_server=None):
        super().__init__(problem)

        if problem.working_dir is None:
            raise Exception('RemoteExecutor: Problem working directory must be set.')

        self.options.declare(name='hostname', default=config["condor_host"],
                             desc='Hostname')
        self.options.declare(name='username', default=config["condor_login"],
                             desc='Username')
        self.options.declare(name='port', default=15900, lower=0,
                             desc='Port')

        # command
        self.command = command

        # files
        self.input_files = files_to_server
        self.output_files = files_from_server

    @abstractmethod
    def eval(self, individual):
        super().eval(individual)

    def _create_client(self):
        # key = os.path.join(os.path.dirname(__file__), "cert/artap.key")
        # cert = os.path.join(os.path.dirname(__file__), "cert/artap.crt")

        channel = rpyc.Channel(rpyc.SocketStream.connect(self.options["hostname"], self.options["port"]))
        channel.send(self.options["username"].encode('utf-8'))
        response = channel.recv()
        AUTH_ERROR = b'error'
        if response == AUTH_ERROR:
            raise rpyc.utils.authenticators.AuthenticationError('Invalid username for daemon')

        return rpyc.utils.factory.connect_channel(channel, service=rpyc.core.ClassicService)
        # return rpyc.classic.connect(self.options["hostname"], self.options["port"])
        # return rpyc.classic.ssl_connect(self.options["hostname"], self.options["port"], keyfile=key, certfile=cert)

    def _init_remote(self, client):
        try:
            # create client
            remote_dir = client.root.create_job_dir()
            return remote_dir
        except RuntimeError as e:
            #self.problem.logger.error("Cannot create remote directory '{}' on ''".format(directory,
            #                                                                             self.options["hostname"]))
            print("ERROR --------------------------------")
            pass

    @staticmethod
    def _create_file_on_remote(destination_file, content, remote_dir, client):
        fp = tempfile.NamedTemporaryFile(mode='w+t', delete=False, newline="\n")
        fp.write(content)
        fp.close()

        RemoteExecutor._transfer_file_to_remote(fp.name, destination_file, remote_dir, client)

        os.unlink(fp.name)

    @staticmethod
    def _transfer_file_to_remote(source_file, destination_file, remote_dir, client):
        source = source_file
        dest = "{}/{}".format(remote_dir, destination_file)

        upload_file(client, localpath=source, remotepath="{}/{}".format(client.root.artap_dir, dest))

    @staticmethod
    def _transfer_file_from_remote(source_file, destination_file, remote_dir, client):
        dest = destination_file
        source = "{}/{}".format(remote_dir, source_file)

        download_file(client, remotepath="{}/{}".format(client.root.artap_dir, source), localpath=dest)

    @staticmethod
    def _create_dir_on_remote(directory, client):
        d = datetime.datetime.now()
        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")

        # projects directory
        client.exec_command("mkdir " + directory)

        # working directory
        wd = "{}/artap-{}".format(directory, ts)
        client.exec_command("mkdir " + wd)

        return wd

    def _transfer_files_to_remote(self, remote_dir, client):
        # transfer input files
        if self.input_files:
            for file in self.input_files:
                self._transfer_file_to_remote("{}/{}".format(self.problem.working_dir, file), "./{}".format(file),
                                              remote_dir=remote_dir, client=client)

    def _transfer_files_from_remote(self, client):
        pass


class CondorJobExecutor(RemoteExecutor):
    """
    Allows distributing of calculation of objective functions.
    """
    def __init__(self, problem, files_to_condor, files_from_condor=None):
        super().__init__(problem, "", files_to_condor, files_from_condor)

        # set default host
        self.options["hostname"] = config["condor_host"]

        # set default requirements
        self.requirements = ""
        self.request_cpus = -1
        self.request_memory = -1
        self.hold_on_start = False

    @abstractmethod
    def _create_job_file(self, remote_dir, individual, client):
        pass

    def _create_desc(self, individual):
        desc = {}
        desc["individual_id"] = individual.id
        desc["individual_population_id"] = individual.population_id
        desc["individual_algorithm_id"] = individual.algorithm_id
        desc["individual_vector"] = '[' + ', '.join(str(e) for e in individual.vector) + ']'

        return desc

    def eval(self, individual):
        if config["condor_host"] is None:
            raise Exception("Condor host is not defined.")

        super().eval(individual)

        success = False
        while not success:
            try:
                # create client
                client = self._create_client()

                # init remote
                remote_dir = self._init_remote(client=client)
                # self.problem.logger.info("RemoteDir {}".format(remote_dir))

                # transfer supplementary files, input and model file
                self._transfer_files_to_remote(remote_dir, client)

                # submit job
                try:
                    self._create_job_file(remote_dir, individual, client)
                except AsyncResultTimeout as e:
                    print("ERROR - {} - try again ({})".format(e, remote_dir))
                    # remove job dir
                    client.root.remove_job_dir(remote_dir)
                    client.close()
                    # resubmit
                    return self.eval(individual)

                start = time.time()

                events = []
                cnt = 0
                return_value = -1
                delay = 0.5

                run = True
                while run:
                    eventlog = client.root.eventlog(remote_dir)
                    for e in eventlog:
                        if {e["timestamp"], e["type"]} not in events:
                            events.append({e["timestamp"], e["type"]})

                    # print("len(events) = {}".format(len(events)))
                    if len(events) > 0:
                        for i in range(cnt, len(events)):
                            event = eventlog[i]
                            tp = event["type"]

                            args = ""

                            if tp == "submit":
                                # SUBMIT
                                if "ReturnValue" in event["info"]:
                                    return_value = event["info"]["ReturnValue"]
                            elif tp == "execute":
                                # EXECUTE
                                if "ExecuteHost" in event["info"]:
                                    args += "ExecuteHost: {}, ".format(parse_address(event["info"]["ExecuteHost"]))
                            elif tp == "image_size":
                                # IMAGE_SIZE
                                pass
                            elif tp == "job_terminated":
                                # JOB_TERMINATED
                                if "ReturnValue" in event["info"]:
                                    return_value = event["info"]["ReturnValue"]
                                run = False
                            elif tp == "job_held":
                                # JOB_HELD
                                if "ReturnValue" in event["info"]:
                                    return_value = event["info"]["ReturnValue"]

                                self.problem.logger.error("Job {}.{} is '{}' at {}".format(event["cluster"], event["proc"], event["type"], ""))
                                run = False
                                # read log
                                #content_log = self._read_file_from_remote("{}.log".format(self.output_files[0]),
                                #                                          remote_dir=remote_dir, client=client)
                                #self.problem.logger.error(content_log)
                                # remove job
                                # self._run_command_on_remote("condor_rm {}".format(process_id),
                                #                            remote_dir=remote_dir, client=client)
                                raise RuntimeError

                            if len(args) > 0:
                                args = args[:-2]

                            self.problem.logger.info("Job {}.{} ({}) is '{}' at {}".format(event["cluster"], event["proc"], remote_dir, event["type"], args))
                            # print("{}: {} ({})".format(eventlog[i].timestamp, eventlog[i].type, args))

                    if run:
                        cnt = len(events)
                        time.sleep(delay)

                end = time.time()
                if (end - start) > self.problem.options["time_out"]:
                    raise TimeoutError

                # successfull
                if return_value == 0:
                    if len(self.output_files) > 0:
                        output_files = []
                        d = datetime.datetime.now()
                        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")
                        path = self.problem.working_dir + 'artap' + ts
                        os.mkdir(path)

                        for file in self.output_files:
                            self._transfer_file_from_remote(source_file=file, destination_file="{}/{}".format(path, file),
                                                            remote_dir=remote_dir, client=client)
                            output_files.append("{}/{}".format(path, file))
                        success = True
                        result = self.parse_results(output_files, individual)
                        # update cost on remote server
                        client.root.log_update_cost(remote_dir, individual, result)

                    # remove job dir
                    if result is not None:
                        client.root.remove_job_dir(remote_dir)

                    if self.problem.options['save_data_files'] is False:
                        self._remove_dir(path)
                else:
                    assert 0

                # remove job dir
                # client.root.remove_job_dir(remote_dir)

                client.close()

                return result

            except ConnectionError as e:
                print(e)
                time.sleep(1.0)
                continue


class CondorPythonJobExecutor(CondorJobExecutor):
    def __init__(self, problem, script, parameter_file, output_files=None, python_path="python3"):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], output_files)

        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

        self.executable = textwrap.dedent("""\
            #!/bin/sh
            # args

            """ + python_path + """  $@""")

        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = self.output_files

        condor_input_files = [self.script]
        # create input file with parameters

        if self.parameter_file:
            # parameters
            # create remote file
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            condor_input_files.append(self.parameter_file)
            arguments = self.script + " " + self.parameter_file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir,
                                        client=client)
        else:
            param_values_string = Executor._join_parameters_values(individual.vector, ",")
            arguments = self.script + " " + param_values_string

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "python"
        desc["extension"] = ".py"
        desc["editor"] = True
        desc["name"] = "Python"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=arguments,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorMatlabJobExecutor(CondorJobExecutor):
    def __init__(self, problem, script, parameter_file, files_from_condor=None, matlab_path="/opt/matlab-R2018b/bin/matlab"):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], files_from_condor)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

        self.executable = textwrap.dedent("""\
            #!/bin/sh
            # args
            s=$@

            """ + matlab_path + """ -nodisplay -nosplash -nodesktop -r ${s%.m}""")

        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.parameter_file, self.input_files[0]]
        condor_output_files = self.output_files

        # create input file with parameters
        if self.parameter_file:
            # parameters
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            # create remote file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir, client=client)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "matlab"
        desc["extension"] = ".m"
        desc["editor"] = False
        desc["name"] = "Matlab"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=self.script,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorComsolJobExecutor(CondorJobExecutor):
    def __init__(self, problem, model_file, files_from_condor=None, comsol_path="/opt/comsol-5.4/bin/comsol"):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

        self.arguments = Template("-inputfile $input_file -nosave -pname $param_names -plist $param_values")
        self.executable = textwrap.dedent("""\
                #!/bin/sh
                """ + comsol_path + """ batch $@
                """)

        self.request_cpus = 3
        self.request_memory = 30
        self.requirements = "(OpSys == \"LINUX\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.model_file]
        condor_output_files = self.output_files

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)
        arguments = self.arguments.substitute(input_file=os.path.basename(self.model_file),
                                              param_names=param_names_string,
                                              param_values=param_values_string)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "comsol"
        desc["extension"] = ".mph"
        desc["editor"] = False
        desc["name"] = "Comsol Multiphysics"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.sh".format(client.root.artap_dir, remote_dir),
                               arguments=arguments,
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


class CondorCSTJobExecutor(CondorJobExecutor):
    def __init__(self, problem, model_file, files_from_condor=None, cst_path="\"C:\Program Files (x86)\CST Studio Suite 2020\CST DESIGN ENVIRONMENT.exe\""):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

        self.executable = textwrap.dedent("""\
            """ + cst_path + """ --se --rebuild --hide -par parameters.txt -project-file {} """.format(self.model_file))
        self.executable += '\n'
        i = 1
        for file in files_from_condor[1:]:
            filename = pathlib.Path(file)
            windows_path = pathlib.PureWindowsPath(filename)
            name = filename.name
            self.output_files[i] = name
            self.executable += 'copy "{}" "{}" \n'.format(windows_path, name)
            i += 1

        # add solver directory to output
        # self.output_files.append(os.path.splitext(self.model_file)[0] + "\Export")

        self.request_cpus = 2
        self.request_memory = 10
        self.requirements = "(OpSys == \"WINDOWS\" && Arch == \"X86_64\")"

    def _create_job_file(self, remote_dir, individual, client):
        condor_input_files = [self.model_file]
        condor_output_files = self.output_files

        parameters = ""
        for parameter, value in zip(self.problem.parameters, individual.vector):
            parameters += "{}={}\n".format(parameter['name'], value)

        parameter_file = "parameters.txt"
        condor_input_files.append(parameter_file)
        self._create_file_on_remote(parameter_file, parameters, remote_dir=remote_dir,
                                    client=client)

        # create executable
        self._create_file_on_remote("run.bat", self.executable, remote_dir=remote_dir, client=client)

        # desc
        desc = self._create_desc(individual)
        desc["type"] = "cst"
        desc["extension"] = ".cst"
        desc["editor"] = False
        desc["name"] = "CST"

        client.root.submit_job(remote_dir=remote_dir,
                               executable="{}/{}/run.bat".format(client.root.artap_dir, remote_dir),
                               arguments="",
                               input_files=condor_input_files,
                               output_files=condor_output_files,
                               requirements=self.requirements,
                               request_cpus=self.request_cpus,
                               request_memory=self.request_memory,
                               hold_on_start=self.hold_on_start,
                               desc=desc)


from .utils import flatten

from .surrogate import SurrogateModel

import numpy as np

from scipy.spatial import distance

from smt.surrogate_models import RBF, KRG, KPLS, IDW, RMTB, LS, QP, GENN, GEKPLS, KPLSK


class SurrogateModelSMT(SurrogateModel):
    def __init__(self, problem):
        super().__init__(problem)

        self.train_step = 10
        self.sigma_threshold = 10
        self.score_threshold = 0.5

        # stats
        self.xxx = None

    @staticmethod
    def get_rbf_regressor(verbose=1):
        # Radial basis functions
        return RBF(d0=5, print_prediction=False)

    @staticmethod
    def get_idw_regressor(verbose=1):
        # Inverse-distance weighting
        return IDW(p=4)

    # @staticmethod
    # def get_rbtb_regressor(verbose=1):
        # Regularized minimal-energy tensor-product splines
        # return RMTB(xlimits=np.array(bench.limits()), order=4, num_ctrl_pts=20, energy_weight=1e-15, regularization_weight=0.0)

    @staticmethod
    def get_ls_regressor(verbose=1):
        # Least-squares approximation
        return LS()

    @staticmethod
    def get_qp_regressor(verbose=1):
        # Second-order polynomial approximation
        return QP()

    @staticmethod
    def get_krg_regressor(verbose=1):
        # Kriging
        return KRG(theta0=[1e-2])

    @staticmethod
    def get_kpls_regressor(verbose=1):
        # Kriging model that uses the partial least squares
        return KPLS(theta0=[1e-2])

    @staticmethod
    def get_kplsk_regressor(verbose=1):
        # Kriging model that uses the partial least squares
        return KPLSK(theta0=[1e-2])

    def init_default_regressor(self):
        # default regressor
        self.regressor = self.get_rbf_regressor()
        self.has_epsilon = False

    def predict(self, x, *args):
        if self.trained:
            return self.regressor.predict_values(np.array([x]))
        else:
            assert 0

    def train(self):
        self.trained = False
        assert(len(self.x_data) == len(self.y_data))

        # print(self.x_data)

        self.regressor.options["print_global"] = False
        self.regressor.set_training_values(np.array(self.x_data), np.array(self.y_data))
        self.regressor.train()

        if self.eval_stats:
            # score
            pass
            # self.score = self.regressor.score(self.x_data, self.y_data)
            # print("self.score = {} : {}".format(len(self.x_data), self.score))
            # lml (Gaussian regressor)
            # if "log_marginal_likelihood" in dir(self.regressor):
            #    self.lml, self.lml_gradient = self.regressor.log_marginal_likelihood(self.regressor.kernel_.theta, eval_gradient=True)

        # set trained
        # self.trained = self.score >= self.score_threshold

        self.trained = True

    def evaluate(self, individual):
        evaluate = True

        if self.trained:
            p = self.predict(individual.vector)
            # REMOVE
            # p_value = self.problem.evaluate(individual)
            # print(individual.vector, p, p_value, abs(p - p_value))
            if self.regressor.supports["variances"]:
                sigma = self.regressor.predict_variances(np.array([individual.vector]))
            else:
                sigma = 0

            # compute minimal parameter distance
            dist = 0.0
            if self.distance_threshold > 0.0:
                dist = self.compute_distance(individual.vector)

                #print("dist = {}".format(dist))
                #print("sigma = {}".format(sigma))

            # estimated value
            value = flatten(p[0])

            # check distance increase counter
            # print("dist = {}".format(dist))
            if dist <= self.distance_threshold:
                self.predict_counter += 1
                evaluate = False

            # HACK
            # if not evaluate:
            #     m = float("inf")
            #     for i in range(len(self.y_data)):
            #         m = min(self.y_data[i][0], m)
            #         # m = min(self.y_data[i], m)
            #     if value[0] < m:
            #     # if value < m:
            #         self.predict_counter -= 1
            #         evaluate = True

        if evaluate:
            # evaluate problem
            value = self.problem.evaluate(individual)
            # increase counter
            self.eval_counter += 1
            # add training date to surrogate model
            self.add_data(individual.vector, value)

            if self.eval_counter % self.train_step == 0:
                # init default regressor
                if self.regressor is None:
                    self.init_default_regressor()

                # train model
                self.train()

        # print("Evaluate = {}, \t value = {}".format(evaluate, value))
        self.problem.logger.info("surrogate: predict / eval counter: {0:5.0f} / {1:5.0f}, total: {2:5.0f}".format(self.problem.surrogate.predict_counter, self.problem.surrogate.eval_counter,
                                                                                                                self.problem.surrogate.predict_counter + self.problem.surrogate.eval_counter))
        return value



import unittest

from ..benchmark_robust import Synthetic1D
from ..algorithm_genetic import EpsMOEA
from ..algorithm import EvaluatorType
from ..results import Results
from ..archive import Archive


# Make test to test something
class TestSimpleOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_worst_case(self):
        # problem = Synthetic1D()
        problem = Synthetic1D()
        algorithm = EpsMOEA(problem, evaluator_type=EvaluatorType.WORST_CASE)
        algorithm.options['max_population_size'] = 50
        algorithm.options['max_population_number'] = 10
        algorithm.run()
        # an archive to collect the eps-dominated solutions
        # archive = Archive(dominance=EpsilonDominance(epsilons=[0.0001, 0.00001]))
        archive = Archive()
        # archive.extend(problem.populations[-1].individuals)
        # archive += problem.populations[-1].individuals

        for individual in problem.last_population():
            archive += individual

            for child in individual.children:
                x_1 = child.vector[0]
                y = child.costs[0]


if __name__ == '__main__':
    unittest.main()




from ..archive import Archive
from ..individual import Individual
from ..operators import crowding_distance

import unittest


class TestArchive(unittest.TestCase):

    def setUp(self):
        self.archive = Archive()

    def test_should_constructor_create_a_non_null_object(self):
        self.assertIsNotNone(self.archive)

    def test_should_adding_one_solution_work_properly(self):
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        self.archive.add(x)
        self.assertEqual(1, self.archive.size())
        self.assertEqual(x, self.archive._contents[0])

    def test_should_adding_two_solutions_work_properly_if_one_is_dominated(self):
        dominated_solution = Individual([1, 2])
        dominated_solution.costs = [2.0, 2.0, 0]
        dominated_solution.costs_signed = [2.0, 2.0, 0]

        dominant_solution = Individual([1, 1])
        dominant_solution.costs = [1.0, 1.0, 0]
        dominant_solution.costs_signed = [1.0, 1.0, 0]

        self.archive.add(dominated_solution)
        self.archive.add(dominant_solution)

        self.assertEqual(1, self.archive.size())
        self.assertEqual(dominant_solution, self.archive._contents[0])

    def test_should_adding_two_solutions_work_properly_if_both_are_non_dominated(self):
        x = Individual([1, 2])
        x.costs = [1.0, 0.0, 0]
        x.costs_signed = [1.0, 0.0, 0]

        y = Individual([1, 1])
        y.costs = [0.0, 1.0, 0]
        y.costs_signed = [0.0, 1.0, 0]

        self.archive.add(x)
        self.archive.add(y)

        self.assertEqual(2, self.archive.size())
        self.assertTrue(x in self.archive._contents and
                        y in self.archive._contents)

    def test_should_adding_four_solutions_work_properly_if_one_dominates_the_others(self):
        x = Individual([1, 2])
        x.costs = [1.0, 1.0]
        x.costs_signed = [1.0, 1.0, 0]

        y = Individual([1, 2])
        y.costs = [0.0, 2.0]
        y.costs_signed = [0.0, 2.0, 0]

        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        v = Individual([1, 2])
        v.costs = [0.0, 0.0]
        v.costs_signed = [0.0, 0.0, 0]

        self.archive.add(x)
        self.archive.add(y)
        self.archive.add(z)
        self.archive.add(v)

        self.assertEqual(1, self.archive.size())
        self.assertEqual(v, self.archive._contents[0])

    def test_should_adding_three_solutions_work_properly_if_two_of_them_are_equal(self):
        x = Individual([1, 2])
        x.costs = [1.0, 1.0]
        x.costs_signed = [1.0, 1.0, 0.0]

        y = Individual([1, 2])
        y.costs = [0.0, 2.0]
        y.costs_signed = [0.0, 2.0, 0.0]

        z = Individual([1, 2])
        z.costs = [1.0, 1.0]
        z.costs_signed = [1.0, 1.0, 0.0]

        self.archive.add(x)
        self.archive.add(y)

        result = self.archive.add(z)

        self.assertEqual(2, self.archive.size())
        self.assertFalse(result)
        self.assertTrue(x in self.archive._contents
                        or y in self.archive._contents)

    def test_crowding_distance_truncate(self):
        # the cost values for the test function, 16 non-dominated solution, half - of- them should be rejected by their
        # crowding distance

        test_costs = [[12, 0],
                      [11.5, 0.5],
                      [11, 1],
                      [10.8, 1.2],
                      [10.5, 1.5],
                      [10.3, 1.8],
                      [9.5, 2],
                      [9, 2.5],
                      [7, 3],
                      [5, 4],
                      [2.5, 6],
                      [2, 10],
                      [1.5, 11],
                      [1, 11.5],
                      [0.8, 11.7],
                      [0, 12]]

        for cost in test_costs:
            x = Individual(cost)
            x.costs = cost
            cost.append(0.)
            x.costs_signed = cost
            x.features = {'crowding_distance': 0}

            res = self.archive.add(x)

        crowding_distance(self.archive._contents)

        # test the algorithm adds every non-dominated solutions to the list
        self.assertEqual(self.archive.size(), len(test_costs))

        # ordering and truncate the elements according to their crowding distance,
        # length of the list should be the given number
        self.archive.truncate(8, 'crowding_distance')
        # if ordered by crowding distance the following items should be contained by the new list
        # the two endpoints
        x1 = Individual([0, 12])
        x1.costs = [0, 12]
        x1.costs_signed = [0, 12, 0.]

        x2 = Individual([12, 0])
        x2.costs = [12, 0]
        x2.costs_signed = [12, 0, 0.]

        self.assertIn(x1, self.archive._contents)
        self.assertIn(x2, self.archive._contents)

        x3 = Individual([1.5, 11])
        x3.costs = [1.5, 11]
        x3.costs_signed = [1.5, 11., 0.]

        self.assertIn(x3, self.archive._contents)



import itertools, operator
from .operators import ParetoDominance, EpsilonDominance
from random import choice, sample


class Archive(object):
    """ Base archiving class based on platypus. An archive only containing non-dominated solutions.
        This base class realize the non-dominated sorting archive.
    """

    def __init__(self, dominance=EpsilonDominance(epsilons=[0.1, 0.1])):
        super(Archive, self).__init__()
        self._dominance = dominance  # dominance comparator
        self._contents = []

    def add(self, individual):
        """
        Archive acceptance procedure is implemented here, this procedure accepts the individual if it is dominates some
        individual and at least non-dominatey by any other member of the list, however it not preserves the length of
        the _contents list.

        :param individual: the examined Individual object from the offsprings.
        :return: 
        """

        is_dominated = False
        is_contained = False

        if len(self._contents) == 0:
            self._contents.append(individual)
            return True
        else:
            number_of_deleted_solutions = 0

            # New copy of list and enumerate
            for index, current_solution in enumerate(list(self._contents)):
                is_dominated_flag = self._dominance.compare(individual.costs_signed, current_solution.costs_signed)
                if is_dominated_flag == 1:
                    del self._contents[index - number_of_deleted_solutions]
                    number_of_deleted_solutions += 1
                elif is_dominated_flag == 2:
                    is_dominated = True
                    break
                elif is_dominated_flag == 0:
                    if individual.costs_signed == current_solution.costs_signed:
                        is_contained = True
                        break

        if not is_dominated and not is_contained:
            self._contents.append(individual)
            return True

        return False

    def rand_choice(self):
        """
        Gives back a random element for the selector operator, here we don't need to
        compare the different levels.
        """
        return choice(self._contents)

    def rand_sample(self, nr=2):
        """Returns an nr long random list of the population """
        return sample(self._contents, nr)

    def truncate(self, size, getter, larger_preferred=True):
        """ Truncates the contents to the given value, which is usually the number of particles/individuals in a
            population. """

        result = sorted(self._contents, key=lambda x: x.features[getter])

        if larger_preferred:
            result.reverse()

        self._contents = result[:size]
        return

    def append(self, individual):
        self.add(individual)

    def extend(self, individuals):
        for individual in individuals:
            self.append(individual)

    def remove(self, solution):
        try:
            self._contents.remove(solution)
            return True
        except ValueError:
            return False

    def size(self) -> int:
        return len(self._contents)

    def __len__(self):
        return len(self._contents)

    def __getitem__(self, key):
        return self._contents[key]

    def __iadd__(self, other):
        if hasattr(other, "__iter__"):
            for o in other:
                self.add(o)
        else:
            self.add(other)

        return self

    def __iter__(self):
        return iter(self._contents)



import unittest

from ..benchmark_functions import Ackley
from ..algorithm_genetic import EpsMOEA
from ..benchmark_pareto import ZDT1
from ..results import Results
from ..quality_indicator import epsilon_add


class TestAckleyN222(unittest.TestCase):
    """ Tests that the eps-moea algorithm can find the global optimum of a function."""

    def test_local_problem(self, population_number=10):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = EpsMOEA(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['epsilons'] = 0.01
            algorithm.options['max_processes'] = 10
            algorithm.run()

            b = Results(problem)
            optimum = b.find_optimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 0)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


class TestZDT1(unittest.TestCase):
    # integration test -- tests the total functionality of eps-MOEA

    def test_local_problem(self):
        problem = problem = ZDT1()
        algorithm = EpsMOEA(problem)
        algorithm.options['max_population_number'] = 500
        algorithm.options['max_population_size'] = 100  # according to the literature
        algorithm.options['max_processes'] = 1
        algorithm.options['epsilons'] = 0.05
        algorithm.run()

        results = Results(problem)
        vals = results.pareto_values(algorithm.archive)
        exact = problem.pareto_front(vals[0])
        self.assertLessEqual(epsilon_add(exact, vals), 0.2)

if __name__ == '__main__':
    unittest.main()



import unittest
from ..benchmark_functions import Ackley
from ..algorithm_swarm import SwarmAlgorithm, OMOPSO
from ..problem import Problem
from ..individual import Individual
from ..archive import Archive


class TestProblem(Problem):
    def set(self):
        self.name = 'Test Problem'
        self.parameters = [{'name': '1', 'bounds': [2., 10.]},
                           {'name': '2', 'bounds': [2., 20.]},
                           ]

        self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        return [1, 1]


class TestSwarm(unittest.TestCase):
    def setUp(self):
        problem = TestProblem()

        self.Swarm = SwarmAlgorithm(problem)

    def test_should_constructor_create_a_valid_object(self):
        self.assertIsNotNone(self.Swarm)

    def test_constrictionfactor(self):
        # results should be 1.
        self.assertEqual(self.Swarm.khi(1., 1.), 1.)

        # result should be -0.2087
        self.assertAlmostEqual(self.Swarm.khi(3., 4.), -0.2087, 4)

    def test_speed_max_should_reduce_speed(self):
        self.assertAlmostEqual(self.Swarm.speed_constriction(1., 1., 0.), 0.5)

    def test_speed_max_should_increase_speed(self):
        self.assertEqual(self.Swarm.speed_constriction(-1., 0., -1.), -0.5)

    def test_speed_do_not_change(self):
        self.assertEqual(self.Swarm.speed_constriction(0.1, 0., -1.), 0.1)

    def test_particle_best_should_initialized_correctly(self):
        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        z.features = {'velocity': [], 'best_cost': [], 'best_vector': []}  #

        population = [z]

        self.Swarm.init_pbest(population)

        self.assertEqual(z.features['best_cost'], z.costs_signed)
        self.assertEqual(z.features['best_vector'], z.vector)

    def test_particle_best_update_should_change_the_values(self):
        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        z.features = {'velocity': [1, 2], 'best_cost': [2., 2., 0.], 'best_vector': [1, 2]}

        population = [z]

        self.Swarm.update_particle_best(population)

        self.assertEqual(z.features['best_cost'], z.costs_signed)
        self.assertEqual(z.features['best_vector'], z.vector)

    def test_particle_best_update_should_not_change_the_values(self):
        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        z.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        population = [z]

        self.Swarm.update_particle_best(population)
        self.assertNotEqual(z.features['best_cost'], z.costs_signed)


class TestOMOPSOTOOLS(unittest.TestCase):
    def setUp(self):
        problem = TestProblem()
        self.omopso = OMOPSO(problem)
        self.omopso.non_uniform_mutator.probability = 1.0
        self.omopso.uniform_mutator.probability = 1.0

    def test_turbulence_should_change_vector(self):
        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        z.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        population = [z]

        self.omopso.turbulence(population)

        # the new population vector should be differ from the original one
        self.assertNotEqual(population[0].vector, [1, 2])

        # it should preserve the features
        self.assertEqual(population[0].features, z.features)

    def test_init_velocity(self):
        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        z.features = {'velocity': [], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        population = [z]

        self.omopso.init_pvelocity(population)
        self.assertEqual(z.features['velocity'], [0, 0])

    def test_global_best(self):
        x = Individual([1, 2])
        x.costs = [2.0, 1.0]
        x.costs_signed = [2.0, 1.0, 0]
        x.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2], 'crowding_distance':[50]}

        self.omopso.leaders.add(x)
        y = self.omopso.select_leader()

        self.assertEqual(x, y)

        z = Individual([3, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]
        z.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2], 'crowding_distance':[100]}

        self.omopso.leaders.add(z)
        y = self.omopso.select_leader()

        # the leader has the bigger crowding distance
        self.assertEqual(z, y)

    def test_velocity(self):
        # if every parameter set to one and we are not on the borders, we shuold got back the velocity
        # mocking the random numbers
        self.omopso.c1_min = 1.
        self.omopso.c1_max = 1.
        self.omopso.c2_min = 1.
        self.omopso.c2_max = 1.0
        self.omopso.r1_min = 1.0
        self.omopso.r1_max = 1.0
        self.omopso.r2_min = 1.0
        self.omopso.r2_max = 1.0
        self.omopso.min_weight = 1.0
        self.omopso.max_weight = 1.0

        x = Individual([1, 2])
        x.costs = [2.0, 1.0]
        x.costs_signed = [2.0, 1.0, 0]

        x.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        # add x elements to the leaders and to the population
        self.omopso.leaders.add(x)
        population = [x]
        self.omopso.update_velocity(population)

        self.assertEqual([1, 2], population[0].features['velocity'])

    def test_position_update(self):
        x = Individual([1, 2])
        x.costs = [2.0, 1.0]
        x.costs_signed = [2.0, 1.0, 0]

        x.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        population = [x]

        self.omopso.update_position(population)

        self.assertEqual(population[0].vector, [2, 4])

    def test_position_update_particle_turn_back_from_the_border(self):
        x = Individual([10, 20])
        x.costs = [10.0, 20.0]
        x.costs_signed = [10.0, 20.0, 0]

        x.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}

        population = [x]

        self.omopso.update_position(population)

        self.assertEqual(population[0].vector, [10, 20])  # preserve the position
        self.assertEqual(population[0].features['velocity'], [-1, -2])  # particle should turn back

    def test_update_global_best(self):
        population = []
        for i in range(0, 10):
            x = Individual([i, 20 - i])
            x.costs = [i, 20.0 - i]
            x.costs_signed = [i, 20.0 - i, 0]

            x.features = {'velocity': [1, 2], 'best_cost': [0., 0., 0.], 'best_vector': [1, 2]}
            population.append(x)

        self.assertIsInstance(self.omopso.archive, Archive)  # archhive is an empty class here

        self.omopso.options['max_population_size'] = 5
        self.omopso.update_global_best(population)

        self.assertEqual(self.omopso.archive.size(), 10)  # archive should contain everybody
        self.assertEqual(self.omopso.leaders.size(), 5)  # shuold be truncated to 5



import unittest
from ..problem import Problem
from ..algorithm_firefly import MoFirefly
from ..results import Results

# import optproblems as optp
# import optproblems.cec2005 as cec2005
#
#
# class CEC2005_TEST_Problems(Problem):
#     """
#     Template class to initialize the different test problems these problems working for 2, 10, 30 variables.
#
#     There are 25 test problems, which were compiled for the Special Session on Real-Parameter Optimization at the
#     Congress on Evolutionary Computation (CEC), Edinburgh, UK, 2-5 Sept. 2005. The mathematical definitions are given in
#     [Suganthan2005]. Additionally, implementations in C, Java, and Matlab were provided for the participants.
#     Artap uses the optproblems test library to run these test. This template class is made to invoke them.
#     """
#
#     def set(self):
#         # two variables
#         self.parameters = [{'name': 'x_1', 'bounds': [func.min_bounds[0], func.max_bounds[0]]},
#                            {'name': 'x_2', 'bounds': [func.min_bounds[1], func.max_bounds[1]]}]
#
#         self.costs = [{'name':'F_1'}]
#
#
#     def evaluate(self, x):
#         problem = optp.Problem(func, num_objectives=1)
#         solutions = [optp.Individual(x.vector)]
#
#         problem.batch_evaluate(solutions)
#
#         return [solutions[0].objective_values]
#
#
# class TestCEC2005(unittest.TestCase):
#     """Tests from CEC2005 library"""
#
#     def run_test_problem(self, nr_gen, function):
#         try:
#             global func
#             func = function
#             problem = CEC2005_TEST_Problems()
#             algorithm = MoFirefly(problem)
#             algorithm.options['max_population_number'] = nr_gen
#             algorithm.options['max_population_size'] = 100
#             algorithm.run()
#
#             b = Results(problem)
#             optimum = b.find_minimum('F_1')  # Takes last cost function
#             self.assertAlmostEqual(optimum.costs[0], func.bias, 0)
#             del func
#         except AssertionError:
#             # try again with more populations
#             max_pop = int(1.5 * nr_gen)
#             # print("Try again with more populations: {}".format(max_pop))
#             self.run_test_problem(max_pop, function)
#
#     def xtest_shifted_sphere(self):
#         self.run_test_problem(50, optp.cec2005.F1(2))
#
#     def xtest_shifted_rot_weierstrass(self):
#         self.run_test_problem(50, optp.cec2005.F11(2))
#
#     def xtest_fletcherpowell(self):  # schwefel
#         self.run_test_problem(50, optp.cec2005.F12(2))
#
#     def xtest_f8f2(self):
#         self.run_test_problem(50, optp.cec2005.F13(2))
#
#     def xtest_expanded_f6(self):
#         self.run_test_problem(50, optp.cec2005.F14(2))


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod
import math


class SurrogateModel(metaclass=ABCMeta):
    def __init__(self, problem):
        # self.name = name
        self.problem = problem

        # surrogate model
        self.regressor = None

        self.x_data = []
        self.y_data = []

        self.trained = False

        # distance in percentage, taking into account bounds
        self.distance_threshold = 0.0

        # stats
        self.eval_counter = 0
        self.predict_counter = 0
        self.eval_stats = True

    def add_data(self, x, y):
        self.x_data.append(x)
        self.y_data.append(y)

    def read_from_data_store(self):
        for population in self.problem.populations:
            for individual in population.individuals:
                self.add_data(individual.vector, individual.costs)

    def init_default_regressor(self):
        pass

    @abstractmethod
    def train(self):
        pass

    @abstractmethod
    def predict(self, x, *args):
        pass

    def compute(self, individual):
        return self.problem.evaluate(individual)

    def compute_distance(self, x):
        # n = len(limits)
        dist = float("inf")
        for vector in self.x_data:
            dist_tmp = 0.0
            for k in range(len(vector)):
                d = 100.0 * abs(vector[k] - x[k]) / (self.problem.parameters[k]['bounds'][1] - self.problem.parameters[k]['bounds'][0])
                # print("parameter {0:11.6e}, \tvector {1:11.6e}, \tdiff {2:11.6e}, \tlimits {3:11.6e} \tdistance: {4:4.1f} %".format(x[k], vector[k], abs(x[k] - vector[k]), self.problem.parameters[k]['bounds'][1] - self.problem.parameters[k]['bounds'][0], d))
                # dist_tmp = dist_tmp + d
                dist_tmp = max(dist_tmp, d)
            # average value
            # dist_tmp = dist_tmp / n

            if dist_tmp < dist:
                vec = vector
            dist = min(dist, dist_tmp)

            # for k in range(len(vector)):
            #     d = 100.0 * math.fabs(vec[k] - x[k]) / (self.problem.parameters[k]['bounds'][1] - self.problem.parameters[k]['bounds'][0])
            #     print("parameter {0:11.6e}, \tvector {1:11.6e}, \tdiff {2:11.6e}, \tlimits {3:11.6e} \tdistance: {4:4.1f} %".format(x[k], vector[k], abs(x[k] - vector[k]), self.problem.parameters[k]['bounds'][1] - self.problem.parameters[k]['bounds'][0], d))
            # print("dist min: {0:4.1f} %".format(dist))

        return dist

    @abstractmethod
    def evaluate(self, x):
        pass


class SurrogateModelEval(SurrogateModel):
    def __init__(self, problem):
        super().__init__(problem)
        self.trained = True

    def train(self):
        pass

    def predict(self, individual):
        return self.problem.evaluate(individual)

    def evaluate(self, individual):
        self.eval_counter += 1
        return self.problem.evaluate(individual)




from .quality_indicator import epsilon_add
from .algorithm_genetic import NSGAII
from .benchmark_pareto import DTLZI, DTLZII, DTLZIII, DTLZIV
from .results import Results
from .individual import Individual

"""
Comparison is based on the following paper:

  [1]  Durillo, Juan J., José García-Nieto, Antonio J. Nebro, Carlos A. Coello Coello, Francisco Luna, and Enrique Alba. 
       "Multi-objective particle swarm optimizers: An experimental comparison."
        In International conference on evolutionary multi-criterion optimization, 
        pp. 495-509. Springer, Berlin, Heidelberg, 2009.

Common parameters
-----------------

    Swarm/Generation size: 100 particles
    Iterations: 250

"""

def b_nsga2_dtlzI():
    """
    Benchmarks the performance of the NSGA-II algorithm with the DTLZI test problem.
    The mean value of the unary epsilon indicator should be around [1]: 7.13e-3  +- 1.6e-3

    :return: the unary epsilon indicator for the dtlz-i test problem
    """

    test2d = DTLZI(**{'dimension': 9, 'm': 2})

    # Cross-check with Platypus and JMetalpy implementatinos
    #print(test2d.evaluate(Individual([1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 0.5])))
    #print(test2d.evaluate(Individual([1.0, 0.1, 0.9, 1.0, 0.8, 0.5, 0.5])))
    #print(test2d.evaluate(Individual([0.6, 0.7, 1.0, 1.0, 1.0, 0.5, 0.5])))

    # expected results:
    #
    # [50.5, 0.0]
    # [33.50000000000001, 0.0]
    # [24.0, 16.0]

    algorithm = NSGAII(test2d)
    algorithm.options['max_population_number'] = 100
    algorithm.options['max_population_size'] = 100
    algorithm.run()

    results = Results(test2d)
    results.pareto_plot()
    vals = results.pareto_values()



    #reference = []
    #for elem in vals:
    #    reference.append((elem[0], 0.5 - elem[0]))

    print(vals)
    #print(reference)

    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    # display the results
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.scatter(vals[0], vals[1], vals[2])
    ax.set_xlim([0, 1.1])
    ax.set_ylim([0, 1.1])
    ax.set_zlim([0, 1.1])
    ax.view_init(elev=30.0, azim=15.0)
    #ax.locator_params(nbins=4)

    plt.show()

b_nsga2_dtlzI()



from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from .operators import GradientEvaluator, RandomGenerator
from .individual import Individual


class GradientDescent(GeneticAlgorithm):
    """ Gradient descent algorithms """

    def __init__(self, problem: Problem, name="GradientDescent"):
        super().__init__(problem, name)
        self.problem = problem

        self.evaluator = GradientEvaluator(self)
        self.options.declare(name='n_iterations', default=10,
                             desc='Maximum number of iterations')
        self.options.declare(name='x0', default=0,
                             desc='initial point')
        self.options.declare(name='h', default=0.1,
                             desc='step')

        self.individual_features['gradient'] = dict()
        self.generator = RandomGenerator(self.problem.parameters, self.individual_features)

    def run(self):
        h = self.options["h"]
        self.generator.init(self.options['max_population_size'])

        # generate
        individuals = self.generator.generate()

        # append to problem
        for individual in individuals:
            self.problem.individuals.append(individual)

        self.evaluate(individuals)

        # TODO: add adaptive step size
        for it in range(self.options['max_population_number']):
            for individual in individuals:
                x = []
                for i in range(len(individual.vector)):
                    x.append(individual.vector[i] - h * individual.features['gradient'][i])
                individual = Individual(x, self.individual_features)
                # append to problem
                self.problem.individuals.append(individual)
                # sync to datastore
                self.problem.data_store.sync_individual(individual)

            self.evaluate(individuals)





from setuptools import Command, setup, find_packages
from setuptools.command.install import install
import os
import datetime


class CustomInstallCommand(install):
    """Custom install setup to help run shell commands (outside shell) before installation"""

    def run(self):
        os.system('python3 -m pip install numpy')
        os.system('python3 -m pip install cython')
        install.run(self)


base_dir = os.path.abspath(os.path.dirname(__file__))

# long description
with open("README.md", "r") as fh:
    long_description = fh.read()

# requirements
with open("requirements.txt", "r") as fh:
    requirements = fh.read().split("\n")

min_numpy_ver = "1.13.3"


def setup_package():
    setuptools_kwargs = {
        "install_requires": [
            f"numpy >= {min_numpy_ver}",
            requirements,
        ],
        "extras_require" : {'full': [f"agrossuite >= 0.01"]},
        "setup_requires": [f"numpy >= {min_numpy_ver}",
                           f"tqdm >= 4.14",
                           f"pkginfo>=1.4.2",
                           f"bleach>=2.1.0"],
        "zip_safe": False,
    }

    dt = datetime.datetime.now()
    
    setup(
        name="artap",
        version="{}.{}".format(dt.strftime('%Y.%m.%d'), dt.hour*60+dt.minute*60+dt.second),
        author=u"Artap Team",
        author_email="artap.framework@gmail.com",
        description="Platform for robust design optimization",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="http://www.agros2d.org/artap/",
        python_requires='>=3.6',
        license="License :: OSI Approved :: MIT License",
        cmdclass={'install': CustomInstallCommand},
        packages=find_packages(),
        include_package_data=True,
        # data_files=[('artap/lib', ['artap/lib/bayesopt.so']),
        # install_requires=requirements,
        # scripts=['3rdparty/submodules.sh'],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Operating System :: POSIX :: Linux",
            "Topic :: Scientific/Engineering",
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        **setuptools_kwargs,
    )


if __name__ == "__main__":
    setup_package()
